"""
define classes to describe AtomSiteAniso
"""
__author__ = 'ikibalin'
__version__ = "2019_08_30"
import os
import numpy


from pycifstar import Global
from cryspy.f_common.cl_fitable import Fitable
from cryspy.f_crystal.cl_adp import ADP

class AtomSiteAniso(object):
    """
    Data items in the ATOM_SITE_ANISO category record details about
    the atom sites in a crystal structure, such as atomic displacement parameters.

    Data items in the ATOM_site_ANISO category record details about
    magnetic properties of the atoms that occupy the atom sites.
    
    Description in cif file:

    loop_
    _atom_site_aniso_label
    _atom_site_aniso_U_11
    _atom_site_aniso_U_22
    _atom_site_aniso_U_33
    _atom_site_aniso_U_12
    _atom_site_aniso_U_13
    _atom_site_aniso_U_23
     O1   .071(1) .076(1) .0342(9) .008(1)   .0051(9) -.0030(9) 
     C2   .060(2) .072(2) .047(1)  .002(2)   .013(1)  -.009(1)  

    
    reference: https://www.iucr.org/__data/iucr/cifdic_html/1/cif_core.dic/Catom_site.html
    """    
    def __init__(self, label=[],  
                 u_11=[], u_22=[], u_33=[],
                 u_12=[], u_13=[], u_23=[]):
        super(AtomSiteAniso, self).__init__()

        self.__atom_site_aniso_label = []
        self.__atom_site_aniso_u_11 = []
        self.__atom_site_aniso_u_12 = []
        self.__atom_site_aniso_u_13 = []
        self.__atom_site_aniso_u_22 = []
        self.__atom_site_aniso_u_23 = []
        self.__atom_site_aniso_u_33 = []

        self.label = label
        self.u_11 = u_11
        self.u_12 = u_12
        self.u_13 = u_13
        self.u_22 = u_22
        self.u_23 = u_23
        self.u_33 = u_33
        
    def __repr__(self):
        ls_out = ["AtomSiteAniso:"]
        ls_out.append(" label    u_11     u_22     u_33     u_12     u_13     u_23    ")
        for hh_1, hh_2, hh_3, hh_4, hh_5, hh_6, hh_7 in zip(
            self.label, self.u_11, self.u_22, self.u_33, self.u_12, self.u_13, self.u_23):
            ls_out.append(" {:8} {:8} {:8} {:8} {:8} {:8} {:8}".format(hh_1, 
                    hh_2.print_with_sigma, hh_3.print_with_sigma, hh_4.print_with_sigma, 
                    hh_5.print_with_sigma, hh_6.print_with_sigma, hh_7.print_with_sigma))
        return "\n".join(ls_out)

    @property
    def label(self):
        """
        Anisotropic atomic displacement parameters are looped in
        a separate list. This code must match the
        _atom_site_label of the associated atom in the atom coordinate
        list and conform with the same rules described in
        _atom_site_label.

        Type: char

        reference: https://www.iucr.org/__data/iucr/cifdic_html/1/cif_core.dic/Iatom_site_aniso_label.html
        """
        return tuple(self.__atom_site_aniso_label)
    @label.setter
    def label(self, l_x):
        l_x_in = []
        for x in l_x:
            x_in = str(x).strip()
            l_x_in.append(x_in)
        self.__atom_site_aniso_label = l_x_in
        len_x = len(l_x_in)

        len_1 = len(self.__atom_site_aniso_u_11)
        if len_1 > len_x:
            self.__atom_site_aniso_u_11 = self.__atom_site_aniso_u_11[:len_x]
        elif len_1 < len_x:
            l_fitable = [Fitable(value=0., name="_atom_site_aniso_u_11") for hh in range(len_x-len_1)]
            self.__atom_site_aniso_u_11.extend(l_fitable)

        len_1 = len(self.__atom_site_aniso_u_22)
        if len_1 > len_x:
            self.__atom_site_aniso_u_22 = self.__atom_site_aniso_u_22[:len_x]
        elif len_1 < len_x:
            l_fitable = [Fitable(value=0., name="_atom_site_aniso_u_22") for hh in range(len_x-len_1)]
            self.__atom_site_aniso_u_22.extend(l_fitable)

        len_1 = len(self.__atom_site_aniso_u_33)
        if len_1 > len_x:
            self.__atom_site_aniso_u_33 = self.__atom_site_aniso_u_33[:len_x]
        elif len_1 < len_x:
            l_fitable = [Fitable(value=0., name="_atom_site_aniso_u_33") for hh in range(len_x-len_1)]
            self.__atom_site_aniso_u_33.extend(l_fitable)

        len_1 = len(self.__atom_site_aniso_u_11)
        if len_1 > len_x:
            self.__atom_site_aniso_u_12 = self.__atom_site_aniso_u_12[:len_x]
        elif len_1 < len_x:
            l_fitable = [Fitable(value=0., name="_atom_site_aniso_u_12") for hh in range(len_x-len_1)]
            self.__atom_site_aniso_u_12.extend(l_fitable)

        len_1 = len(self.__atom_site_aniso_u_13)
        if len_1 > len_x:
            self.__atom_site_aniso_u_13 = self.__atom_site_aniso_u_13[:len_x]
        elif len_1 < len_x:
            l_fitable = [Fitable(value=0., name="_atom_site_aniso_u_13") for hh in range(len_x-len_1)]
            self.__atom_site_aniso_u_13.extend(l_fitable)

        len_1 = len(self.__atom_site_aniso_u_23)
        if len_1 > len_x:
            self.__atom_site_aniso_u_23 = self.__atom_site_aniso_u_23[:len_x]
        elif len_1 < len_x:
            l_fitable = [Fitable(value=0., name="_atom_site_aniso_u_23") for hh in range(len_x-len_1)]
            self.__atom_site_aniso_u_23.extend(l_fitable)

    @property
    def u_11(self):
        """
        These are the standard anisotropic atomic displacement
        components in angstroms squared which appear in the
        structure-factor term

        T = exp{-2pi^2^ sum~i~ [sum~j~ (U^ij^ h~i~ h~j~ a*~i~ a*~j~) ] }

        h = the Miller indices
        a* = the reciprocal-space cell lengths

        The unique elements of the real symmetric matrix are
        entered by row.

        Default: 0.

        Type: float

        reference: https://www.iucr.org/__data/iucr/cifdic_html/1/cif_core.dic/Iatom_site_aniso_U_.html
        """
        return tuple(self.__atom_site_aniso_u_11)
    @u_11.setter
    def u_11(self, l_x):
        l_fitable = []
        for x in l_x:
            if isinstance(x, Fitable):
                x_in = x
            else:
                x_in = Fitable()
                flag = x_in.take_it(x)
            l_fitable.append(x_in)
        len_x = len(l_fitable)
        len_1 = len(self.__atom_site_aniso_label)
        if len_1 < len_x:
            l_fitable = l_fitable[:len_1]
        elif len_1 > len_x:
            l_fitable.extend([Fitable(value=0., name= "_atom_site_aniso_u_11") for hh in range(len_1-len_x)])
        self.__atom_site_aniso_u_11 = l_fitable


    @property
    def u_22(self):
        """
        see help for u_11 parameter
        """
        return tuple(self.__atom_site_aniso_u_22)
    @u_22.setter
    def u_22(self, l_x):
        l_fitable = []
        for x in l_x:
            if isinstance(x, Fitable):
                x_in = x
            else:
                x_in = Fitable()
                flag = x_in.take_it(x)
            l_fitable.append(x_in)
        len_x = len(l_fitable)
        len_1 = len(self.__atom_site_aniso_label)
        if len_1 < len_x:
            l_fitable = l_fitable[:len_1]
        elif len_1 > len_x:
            l_fitable.extend([Fitable(value=0., name= "_atom_site_aniso_u_22") for hh in range(len_1-len_x)])
        self.__atom_site_aniso_u_22 = l_fitable


    @property
    def u_33(self):
        """
        see help for u_11 parameter
        """
        return tuple(self.__atom_site_aniso_u_33)
    @u_33.setter
    def u_33(self, l_x):
        l_fitable = []
        for x in l_x:
            if isinstance(x, Fitable):
                x_in = x
            else:
                x_in = Fitable()
                flag = x_in.take_it(x)
            l_fitable.append(x_in)
        len_x = len(l_fitable)
        len_1 = len(self.__atom_site_aniso_label)
        if len_1 < len_x:
            l_fitable = l_fitable[:len_1]
        elif len_1 > len_x:
            l_fitable.extend([Fitable(value=0., name= "_atom_site_aniso_u_33") for hh in range(len_1-len_x)])
        self.__atom_site_aniso_u_33 = l_fitable


    @property
    def u_12(self):
        """
        see help for u_11 parameter
        """
        return tuple(self.__atom_site_aniso_u_12)
    @u_12.setter
    def u_12(self, l_x):
        l_fitable = []
        for x in l_x:
            if isinstance(x, Fitable):
                x_in = x
            else:
                x_in = Fitable()
                flag = x_in.take_it(x)
            l_fitable.append(x_in)
        len_x = len(l_fitable)
        len_1 = len(self.__atom_site_aniso_label)
        if len_1 < len_x:
            l_fitable = l_fitable[:len_1]
        elif len_1 > len_x:
            l_fitable.extend([Fitable(value=0., name= "_atom_site_aniso_u_12") for hh in range(len_1-len_x)])
        self.__atom_site_aniso_u_12 = l_fitable


    @property
    def u_13(self):
        """
        see help for u_11 parameter
        """
        return tuple(self.__atom_site_aniso_u_13)
    @u_13.setter
    def u_13(self, l_x):
        l_fitable = []
        for x in l_x:
            if isinstance(x, Fitable):
                x_in = x
            else:
                x_in = Fitable()
                flag = x_in.take_it(x)
            l_fitable.append(x_in)
        len_x = len(l_fitable)
        len_1 = len(self.__atom_site_aniso_label)
        if len_1 < len_x:
            l_fitable = l_fitable[:len_1]
        elif len_1 > len_x:
            l_fitable.extend([Fitable(value=0., name= "_atom_site_aniso_u_13") for hh in range(len_1-len_x)])
        self.__atom_site_aniso_u_13 = l_fitable


    @property
    def u_23(self):
        """
        see help for u_11 parameter
        """
        return tuple(self.__atom_site_aniso_u_23)
    @u_23.setter
    def u_23(self, l_x):
        l_fitable = []
        for x in l_x:
            if isinstance(x, Fitable):
                x_in = x
            else:
                x_in = Fitable()
                flag = x_in.take_it(x)
            l_fitable.append(x_in)
        len_x = len(l_fitable)
        len_1 = len(self.__atom_site_aniso_label)
        if len_1 < len_x:
            l_fitable = l_fitable[:len_1]
        elif len_1 > len_x:
            l_fitable.extend([Fitable(value=0., name= "_atom_site_aniso_u_23") for hh in range(len_1-len_x)])
        self.__atom_site_aniso_u_23 = l_fitable

    def _show_message(self, s_out: str):
        print("***  Error ***")
        print(s_out)

    @property
    def is_defined(self):
        """
        Output: True if all started parameters are given
        """
        cond = (self.label != [])
        return cond
    @property
    def is_variable(self):
        res = (any([hh.refinement for hh in self.u_11]) | 
               any([hh.refinement for hh in self.u_22]) |
               any([hh.refinement for hh in self.u_33]) |
               any([hh.refinement for hh in self.u_12]) |
               any([hh.refinement for hh in self.u_13]) |
               any([hh.refinement for hh in self.u_23]))
        return res


    def get_variables(self):
        l_variable = [hh for hh in self.u_11 if hh.refinement]
        l_variable.extend([hh for hh in self.u_22 if hh.refinement])
        l_variable.extend([hh for hh in self.u_33 if hh.refinement])
        l_variable.extend([hh for hh in self.u_12 if hh.refinement])
        l_variable.extend([hh for hh in self.u_13 if hh.refinement])
        l_variable.extend([hh for hh in self.u_23 if hh.refinement])

        return l_variable

    @property
    def to_cif(self):
        ls_out = []
        if self.is_defined:
            ls_out.append("loop_")
            ls_out.append("_atom_site_aniso_label")
            ls_out.append("_atom_site_aniso_u_11")
            ls_out.append("_atom_site_aniso_u_22")
            ls_out.append("_atom_site_aniso_u_33")
            ls_out.append("_atom_site_aniso_u_12")
            ls_out.append("_atom_site_aniso_u_13")
            ls_out.append("_atom_site_aniso_u_23")
            for hh_1, hh_2, hh_3, hh_4, hh_5, hh_6, hh_7 in zip(self.label, 
                self.u_11, self.u_22, self.u_33, self.u_12, self.u_13, self.u_23):
                ls_out.append("{:} {:} {:} {:} {:} {:} {:}".format(hh_1, 
                        hh_2.print_with_sigma, hh_3.print_with_sigma, hh_4.print_with_sigma, 
                        hh_5.print_with_sigma, hh_6.print_with_sigma, hh_7.print_with_sigma))
        return "\n".join(ls_out)

    def from_cif(self, string: str):
        cif_global = Global()
        flag = cif_global.take_from_string(string)
        if not flag:
            return False
        flag = False
        flag = cif_global.is_prefix("_atom_site_aniso")
        if flag:
            cif_loop = cif_global["_atom_site_aniso"]
            l_name = cif_loop.names
            if "_atom_site_aniso_label" in l_name:
                self.label = cif_loop["_atom_site_aniso_label"]

            if "_atom_site_aniso_u_11" in l_name:
                l_val = cif_loop["_atom_site_aniso_u_11"]
                l_fitable = []
                for val in l_val:
                    fitable = Fitable(name="_atom_site_aniso_u_11")
                    fitable.take_it(val)
                    l_fitable.append(fitable)
                self.u_11 = l_fitable
            if "_atom_site_aniso_u_22" in l_name:
                l_val = cif_loop["_atom_site_aniso_u_22"]
                l_fitable = []
                for val in l_val:
                    fitable = Fitable(name="_atom_site_aniso_u_22")
                    fitable.take_it(val)
                    l_fitable.append(fitable)
                self.u_22 = l_fitable
            if "_atom_site_aniso_u_33" in l_name:
                l_val = cif_loop["_atom_site_aniso_u_33"]
                l_fitable = []
                for val in l_val:
                    fitable = Fitable(name="_atom_site_aniso_u_33")
                    fitable.take_it(val)
                    l_fitable.append(fitable)
                self.u_33 = l_fitable
            if "_atom_site_aniso_u_12" in l_name:
                l_val = cif_loop["_atom_site_aniso_u_12"]
                l_fitable = []
                for val in l_val:
                    fitable = Fitable(name="_atom_site_aniso_u_12")
                    fitable.take_it(val)
                    l_fitable.append(fitable)
                self.u_12 = l_fitable
            if "_atom_site_aniso_u_13" in l_name:
                l_val = cif_loop["_atom_site_aniso_u_13"]
                l_fitable = []
                for val in l_val:
                    fitable = Fitable(name="_atom_site_aniso_u_13")
                    fitable.take_it(val)
                    l_fitable.append(fitable)
                self.u_13 = l_fitable
            if "_atom_site_aniso_u_23" in l_name:
                l_val = cif_loop["_atom_site_aniso_u_23"]
                l_fitable = []
                for val in l_val:
                    fitable = Fitable(name="_atom_site_aniso_u_23")
                    fitable.take_it(val)
                    l_fitable.append(fitable)
                self.u_23 = l_fitable
        else:
            self.label = []

        return True

    def _form_adp(self, atom_site):
        label = numpy.array(atom_site.label, dtype=str)
        b_iso = numpy.array(atom_site.b_iso, dtype=float)
        adp_type = numpy.array(atom_site.adp_type, dtype=str)

        b_iso_in = numpy.zeros(b_iso.shape, dtype=float)
        b_iso_in = numpy.where(adp_type=="uiso", b_iso, b_iso_in)
        label_aniso = numpy.array(self.label, dtype=str)
        if not(set(label_aniso).issubset(set(label))):
            self._show_message("Unknown 'aniso_label'")
            return False
        
        u_11 = numpy.array(self.u_11, dtype=float)
        u_22 = numpy.array(self.u_22, dtype=float)
        u_33 = numpy.array(self.u_33, dtype=float)
        u_12 = numpy.array(self.u_12, dtype=float)
        u_13 = numpy.array(self.u_13, dtype=float)
        u_23 = numpy.array(self.u_23, dtype=float)


        np_index = numpy.array([int(numpy.argwhere(label==hh)[0]) for hh in label_aniso], dtype=int)

        u_11_in = numpy.zeros(label.shape, dtype=float)
        u_22_in = numpy.zeros(label.shape, dtype=float)
        u_33_in = numpy.zeros(label.shape, dtype=float)
        u_12_in = numpy.zeros(label.shape, dtype=float)
        u_13_in = numpy.zeros(label.shape, dtype=float)
        u_23_in = numpy.zeros(label.shape, dtype=float)

        u_11_in[np_index], u_22_in[np_index], u_33_in[np_index] = u_11, u_22, u_33
        u_12_in[np_index], u_13_in[np_index], u_23_in[np_index] = u_12, u_13, u_23

        np_zeros = numpy.zeros(label.shape, dtype=float)
        u_11_in = numpy.where(adp_type=="uani", u_11_in, np_zeros)
        u_22_in = numpy.where(adp_type=="uani", u_22_in, np_zeros)
        u_33_in = numpy.where(adp_type=="uani", u_33_in, np_zeros)
        u_12_in = numpy.where(adp_type=="uani", u_12_in, np_zeros)
        u_13_in = numpy.where(adp_type=="uani", u_13_in, np_zeros)
        u_23_in = numpy.where(adp_type=="uani", u_23_in, np_zeros)

        adp = ADP(u_11=u_11_in, u_22=u_22_in, u_33=u_33_in, 
                  u_12=u_12_in, u_13=u_13_in, u_23=u_23_in, 
                  b_iso=b_iso_in)
        return adp

    def apply_space_group_constraint(self, atom_site, space_group):
        """
        according to table 1 in Peterse, Palm, Acta Cryst.(1966), 20, 147
        """
        l_numb = atom_site.calc_constr_number(space_group)
        label_aniso = self.label
        label = atom_site.label
        l_ind = [label.index(_1) for _1 in label_aniso]
        for index, u_11, u_22, u_33, u_12, u_13, u_23 in zip(l_ind, self.u_11, self.u_22, self.u_33, 
                                                                    self.u_12, self.u_13, self.u_23):
            u_11.constraint_flag, u_22.constraint_flag, u_33.constraint_flag = False, False, False
            u_12.constraint_flag, u_13.constraint_flag, u_23.constraint_flag = False, False, False
            numb = l_numb[index]
            if numb == 1:
                u_12.value = 0.
                u_12.refinement = False
                u_12.constraint_flag = True
                u_23.value = 0.
                u_23.refinement = False
                u_23.constraint_flag = True
            elif numb == 2:
                u_23.value = 0.
                u_23.refinement = False
                u_23.constraint_flag = True
                u_13.value = 0.
                u_13.refinement = False
                u_13.constraint_flag = True
            elif numb == 3:
                u_12.value = 0.
                u_12.refinement = False
                u_12.constraint_flag = True
                u_13.value = 0.
                u_13.refinement = False
                u_13.constraint_flag = True
            elif numb == 4:
                u_12.value = 0.
                u_12.refinement = False
                u_12.constraint_flag = True
                u_13.value = 0.
                u_13.refinement = False
                u_13.constraint_flag = True
                u_23.value = 0.
                u_23.refinement = False
                u_23.constraint_flag = True
            elif numb == 5:
                u_22.value = u_11.value
                u_22.refinement = False
                u_22.constraint_flag = True
                u_13.value = 0.
                u_13.refinement = False
                u_13.constraint_flag = True
                u_23.value = 0.
                u_23.refinement = False
                u_23.constraint_flag = True
            elif numb == 6:
                u_22.value = u_11.value
                u_22.refinement = False
                u_22.constraint_flag = True
                u_23.value = u_13.value 
                u_23.refinement = False
                u_23.constraint_flag = True
            elif numb == 7:
                u_22.value = u_11.value
                u_22.refinement = False
                u_22.constraint_flag = True
                u_23.value = -1.*u_13.value 
                u_23.refinement = False
                u_23.constraint_flag = True
            elif numb == 8:
                u_22.value = u_11.value
                u_22.refinement = False
                u_22.constraint_flag = True
                u_12.value = 0.
                u_12.refinement = False
                u_12.constraint_flag = True
                u_13.value = 0.
                u_13.refinement = False
                u_13.constraint_flag = True
                u_23.value = 0.
                u_23.refinement = False
                u_23.constraint_flag = True
            elif numb == 9:
                u_33.value = u_22.value
                u_33.refinement = False
                u_33.constraint_flag = True
                u_12.value = 0.
                u_12.refinement = False
                u_12.constraint_flag = True
                u_13.value = 0.
                u_13.refinement = False
                u_13.constraint_flag = True
            elif numb == 10:
                u_33.value = u_22.value
                u_33.refinement = False
                u_33.constraint_flag = True
                u_13.value = 1.*u_12.value 
                u_13.refinement = False
                u_13.constraint_flag = True
            elif numb == 11:
                u_33.value = u_22.value
                u_33.refinement = False
                u_33.constraint_flag = True
                u_13.value = -1.*u_12.value 
                u_13.refinement = False
                u_13.constraint_flag = True
            elif numb == 12:
                u_33.value = u_22.value
                u_33.refinement = False
                u_33.constraint_flag = True
                u_12.value = 0.
                u_12.refinement = False
                u_12.constraint_flag = True
                u_13.value = 0.
                u_13.refinement = False
                u_13.constraint_flag = True
                u_23.value = 0.
                u_23.refinement = False
                u_23.constraint_flag = True
            elif numb == 13:
                u_12.value = 0.5*u_22.value
                u_12.refinement = False
                u_12.constraint_flag = True
                u_23.value = 0.
                u_23.refinement = False
                u_23.constraint_flag = True
            elif numb == 14:
                u_12.value = 0.5*u_22.value
                u_12.refinement = False
                u_12.constraint_flag = True
                u_13.value = 0.
                u_13.refinement = False
                u_13.constraint_flag = True
                u_23.value = 0.
                u_23.refinement = False
                u_23.constraint_flag = True
            elif numb == 15:
                u_12.value = 0.5*u_22.value
                u_12.refinement = False
                u_12.constraint_flag = True
                u_23.value = 2.*u_13.value
                u_23.refinement = False
                u_23.constraint_flag = True
            elif numb == 16:
                u_22.value = 1.0*u_11.value
                u_22.refinement = False
                u_22.constraint_flag = True
                u_12.value = 0.5*u_11.value
                u_12.refinement = False
                u_12.constraint_flag = True
                u_13.value = 0.
                u_13.refinement = False
                u_13.constraint_flag = True
                u_23.value = 0.
                u_23.refinement = False
                u_23.constraint_flag = True
            elif numb == 17:
                u_22.value = 1.0*u_11.value
                u_22.refinement = False
                u_22.constraint_flag = True
                u_33.value = 1.0*u_11.value
                u_33.refinement = False
                u_33.constraint_flag = True
                u_12.value = 0.
                u_12.refinement = False
                u_12.constraint_flag = True
                u_13.value = 0.
                u_13.refinement = False
                u_13.constraint_flag = True
                u_23.value = 0.
                u_23.refinement = False
                u_23.constraint_flag = True
            elif numb == 18:
                u_22.value = 1.0*u_11.value
                u_22.refinement = False
                u_22.constraint_flag = True
                u_33.value = 1.0*u_11.value
                u_33.refinement = False
                u_33.constraint_flag = True
                u_13.value = 1.0*u_12.value
                u_13.refinement = False
                u_13.constraint_flag = True
                u_23.value = 1.0*u_12.value
                u_23.refinement = False
                u_23.constraint_flag = True


    def calc_beta(self, cell):
        """
        calculate beta_ij from U_ij
        """
        ia, ib, ic = cell.ia, cell.ib, cell.ic
        u_11, u_22, u_33 = numpy.array(self.u_11, float), numpy.array(self.u_22, float), numpy.array(self.u_33, float)
        u_12, u_13, u_23 = numpy.array(self.u_12, float), numpy.array(self.u_13, float), numpy.array(self.u_23, float)
        beta_11 = 2.*numpy.pi**2*u_11*ia**2
        beta_22 = 2.*numpy.pi**2*u_22*ib**2
        beta_33 = 2.*numpy.pi**2*u_33*ic**2
        beta_12 = 2.*numpy.pi**2*u_12*ia*ib
        beta_13 = 2.*numpy.pi**2*u_13*ia*ic
        beta_23 = 2.*numpy.pi**2*u_23*ib*ic
        return beta_11, beta_22, beta_33, beta_12, beta_13, beta_23


# -*- coding: utf-8 -*-
"""
Simple RhoChi viewer
"""
__author__ = 'ikibalin'
__version__ = "2019_09_10"

import os
import os.path
import sys
import numpy

import pycifstar

f_dir = os.path.dirname(__file__)
os.chdir(f_dir)
sys.path.insert(0, f_dir)

from cryspy import RhoChi, rhochi_refinement
from cryspy.scripts.rhochi.cl_rhochi import create_temporary

from PyQt5 import QtWidgets
from PyQt5 import QtGui
from PyQt5 import QtCore


import interactive_graph_mod_mono
import interactive_graph_mod_pwd
import interactive_graph_mod_pwd_2d_as_1d



def del_layout(layout):
    """
    delete all elements from layouts
    """
    for i in reversed(range(layout.count())):
        if layout.itemAt(i).widget() != None:
           layout.itemAt(i).widget().setParent(None)
        elif layout.itemAt(i).layout() != None:
            del_layout(layout.itemAt(i).layout())
            layout.itemAt(i).layout().setParent(None)
        else:
            layout.removeItem(layout.itemAt(i))
    return

def multistring_to_array(string):
    l_1 = string.strip().split("\n")
    l_ttheta = [float(_) for _ in l_1[0].strip().split()[1:]]
    l_phi, ll_intensity = [], []
    for line in l_1[1:]:
        l_1 = line.strip().split()
        l_phi.append(float(l_1[0]))
        ll_intensity.append([float(_) if _ != "None" else None for _ in l_1[1:]])
    x = numpy.array(l_ttheta, dtype=float)
    y = numpy.array(l_phi, dtype=float)
    zz = (numpy.array(ll_intensity, dtype=float)).transpose()
    return x, y, zz


class mythread(QtCore.QThread):
    def __init__(self, parent=None):
        QtCore.QThread.__init__(self,parent)
        self.signal = None
        self.core = None
        #self.startpatam = startparam

    def run(self):
        ccore = self.core
        try:
            lmessage = ccore.run_refinement()
            model = cmodel.cmodel()
            model.take_from_core(ccore)
            self.signal.model = model
        except:
            lmessage = ["Undefined mistake during the refinement procedure.\nSomething wrong."]
        print("Calculations are stopped")
        self.signal.message = lmessage
        self.signal.rename_signal.emit()


class cthread_signal(QtCore.QObject):
    rename_signal = QtCore.pyqtSignal()


class cbuilder(QtWidgets.QMainWindow):
    def __init__(self, f_dir_prog = os.path.dirname(__file__)):
        super(cbuilder, self).__init__()
        self._p_f_dir_prog = f_dir_prog 
        self._p_d_setup = {}

        self.read_setup()
        self.def_actions()
        self.init_widget()

        #self.setStyleSheet("background-color:lightgray;")
        self.show()

    def read_setup(self):
        f_dir_prog = self._p_f_dir_prog
        f_file_prog_setup = os.path.join(f_dir_prog, 'setup.dat')

        if not(os.path.isfile(f_file_prog_setup)):
            self.write_setup()

        fid = open(f_file_prog_setup, "r")
        l_cont = fid.readlines()
        fid.close()
        d_setup = {}
        for line in l_cont:
            l_help = [hh.strip() for hh in line.strip().split("::")]
            if len(l_help) == 2:
                d_setup.update({l_help[0]: l_help[1]})
        
        self._p_d_setup = d_setup
        
        if "f_dir_data" in d_setup.keys():
            self.setWindowTitle('RhoChi Viewer - {:}'.format(d_setup["f_dir_data"]))

    def write_setup(self):
        l_key = ["f_dir_data"]
        l_key_val = [self._p_f_dir_prog]

        d_setup = self._p_d_setup

        if (("f_name_data" in d_setup.keys()) & ("f_dir_data" in d_setup.keys())):
            self.setWindowTitle('RhoChi Viewer - {:}'.format(d_setup["f_dir_data"]))


        l_key_setup = d_setup.keys()
        for key, key_val in zip(l_key, l_key_val):
            if key not in l_key_setup:
                d_setup.update({key: key_val})
        ls_out = ["{:}::{:}".format(key, val) for key, val in d_setup.items()]

        f_dir_prog = self._p_f_dir_prog
        f_file_prog_setup = os.path.join(f_dir_prog, 'setup.dat')
        fid = open(f_file_prog_setup, "w")
        fid.write("\n".join(ls_out))
        fid.close()


    def def_actions(self):

        f_dir_prog = self._p_f_dir_prog
        f_dir_prog_icon = os.path.join(f_dir_prog, 'f_icon')

        create_rcif_action = QtWidgets.QAction(QtGui.QIcon(os.path.join(f_dir_prog_icon,'create_new.png')),'Create &new', self)
        create_rcif_action.setShortcut('Ctrl+N')
        create_rcif_action.setStatusTip('Create new rcif')
        create_rcif_action.triggered.connect(self.create_rcif)

        open_rcif_action = QtWidgets.QAction(QtGui.QIcon(os.path.join(f_dir_prog_icon,'open.png')),'&Open .rcif', self)
        open_rcif_action.setShortcut('Ctrl+O')
        open_rcif_action.setStatusTip('Open data (1D)')
        open_rcif_action.triggered.connect(self.open_rcif)


        run_rhochi = QtWidgets.QAction(QtGui.QIcon(os.path.join(f_dir_prog_icon,'calc_rcif.png')),'&Run RhoChi', self)
        run_rhochi.setShortcut('Ctrl+R')
        run_rhochi.setStatusTip('Model optimization of marked parameters')
        run_rhochi.triggered.connect(self.run_rhochi)

        read_rcif_action = QtWidgets.QAction(QtGui.QIcon(os.path.join(f_dir_prog_icon,'read_rcif.png')),'Read .rcif', self)
        read_rcif_action.setStatusTip('Read .rcif')
        read_rcif_action.triggered.connect(self.read_rcif)



        self.statusBar()

        menubar = self.menuBar()
        fileMenu = menubar.addMenu('&File')
        fileMenu.addAction(create_rcif_action)
        fileMenu.addAction(open_rcif_action)

        calcMenu = menubar.addMenu('&Calculations')
        calcMenu.addAction(run_rhochi)
        calcMenu.addAction(read_rcif_action)


        self.toolbar = self.addToolBar("Open")
        self.toolbar.addAction(create_rcif_action)
        self.toolbar.addAction(open_rcif_action)
        self.toolbar.addAction(run_rhochi)
        self.toolbar.addAction(read_rcif_action)

    def init_widget(self): 
        """
        module is reloaded for cwidget
        """
        self.location_on_the_screen()
        self._p_widg_central = cwidget(self.info_width)

        self.setCentralWidget(self._p_widg_central)
        self.try_to_plot_from_rcif()

    def try_to_plot_from_rcif(self):
        try:
            f_name_data = self._p_d_setup["f_name_data"]
            f_dir_data = self._p_d_setup["f_dir_data"]
            if f_name_data == "main.rcif":
                l_file = [_ for _ in os.listdir(f_dir_data) if os.path.isfile(os.path.join(f_dir_data, _))]
                l_file_calc = [_ for _ in l_file if _.endswith("_calc.rcif")]
                if len(l_file_calc) != 0:
                    f_name_data = l_file_calc[0]
            f_name_full = os.path.join(f_dir_data, f_name_data)
            cif_global = pycifstar.to_global(f_name_full)

            flag_single = cif_global.is_prefix("_diffrn_refln")
            flag_powder_1d = cif_global.is_prefix("_pd_proc")
            flag_powder_2d = cif_global.is_prefix("_pd2d_proc")
            if flag_single:
                x = numpy.array(cif_global["_diffrn_refln_fr_calc"], dtype=float)
                y_exp = numpy.array(cif_global["_diffrn_refln_fr"], dtype=float)
                y_sig = numpy.array(cif_global["_diffrn_refln_fr_sigma"], dtype=float)
                self.plot_data_mono(x, y_exp, y_sig)
            if flag_powder_1d:
                x = numpy.array(cif_global["_pd_proc_2theta"], dtype=float)
                y_1_mod = numpy.array(cif_global["_pd_proc_intensity_up_total"], dtype=float)
                y_2_mod = numpy.array(cif_global["_pd_proc_intensity_down_total"], dtype=float)
                y_1_exp = numpy.array(cif_global["_pd_proc_intensity_up"], dtype=float)
                y_1_sig = numpy.array(cif_global["_pd_proc_intensity_up_sigma"], dtype=float)
                y_2_exp = numpy.array(cif_global["_pd_proc_intensity_down"], dtype=float)
                y_2_sig = numpy.array(cif_global["_pd_proc_intensity_down_sigma"], dtype=float)

                y_3_mod = y_1_mod + y_2_mod
                y_4_mod = y_1_mod - y_2_mod
                y_3_exp = y_1_exp + y_2_exp
                y_4_exp = y_1_exp - y_2_exp
                y_3_sig = (y_1_sig**2 + y_2_sig**2)**0.5
                y_4_sig = (y_1_sig**2 + y_2_sig**2)**0.5

                l_y_mod = [y_3_mod, y_4_mod]
                l_y_exp = [y_3_exp, y_4_exp]
                l_y_sig = [y_3_sig, y_4_sig]
                self.plot_data_powder_1d(x, l_y_mod, l_y_exp, l_y_sig)
            if flag_powder_2d:
                s_1_mod = cif_global["_pd2d_proc_2theta_phi_intensity_up_total"].value
                s_2_mod = cif_global["_pd2d_proc_2theta_phi_intensity_down_total"].value
                s_1_exp = cif_global["_pd2d_proc_2theta_phi_intensity_up"].value
                s_1_sig = cif_global["_pd2d_proc_2theta_phi_intensity_up_sigma"].value
                s_2_exp = cif_global["_pd2d_proc_2theta_phi_intensity_down"].value
                s_2_sig = cif_global["_pd2d_proc_2theta_phi_intensity_down_sigma"].value
                x, _2, yy_1_mod = multistring_to_array(s_1_mod)
                x, _2, yy_2_mod = multistring_to_array(s_2_mod)
                x, _2, yy_1_exp = multistring_to_array(s_1_exp)
                x, _2, yy_2_exp = multistring_to_array(s_2_exp)
                x, _2, yy_1_sig = multistring_to_array(s_1_sig)
                x, _2, yy_2_sig = multistring_to_array(s_2_sig)
                y_1_exp = numpy.where(numpy.isnan(yy_1_exp), 0., yy_1_exp).sum(axis=1)
                y_1_mod = numpy.where(numpy.isnan(yy_1_exp), 0., yy_1_mod).sum(axis=1)
                y_1_sig = ((numpy.where(numpy.isnan(yy_1_exp), 0., yy_1_sig)**2).sum(axis=1))**0.5

                y_2_exp = numpy.where(numpy.isnan(yy_2_exp), 0., yy_2_exp).sum(axis=1)
                y_2_mod = numpy.where(numpy.isnan(yy_2_exp), 0., yy_2_mod).sum(axis=1)
                y_2_sig = ((numpy.where(numpy.isnan(yy_2_exp), 0., yy_2_sig)**2).sum(axis=1))**0.5

                y_3_mod = y_1_mod + y_2_mod
                y_4_mod = y_1_mod - y_2_mod
                y_3_exp = y_1_exp + y_2_exp
                y_4_exp = y_1_exp - y_2_exp
                y_3_sig = (y_1_sig**2 + y_2_sig**2)**0.5
                y_4_sig = (y_1_sig**2 + y_2_sig**2)**0.5

                l_y_mod = [y_3_mod, y_4_mod]
                l_y_exp = [y_3_exp, y_4_exp]
                l_y_sig = [y_3_sig, y_4_sig]
                self.plot_data_powder_1d(x, l_y_mod, l_y_exp, l_y_sig)
        except:
            pass

    def location_on_the_screen(self):
        """
        position and size of main window
        """
        screen = QtWidgets.QDesktopWidget().screenGeometry()
        self.setMinimumSize(screen.width()*1/4, screen.height()*1/4)
        #self.setMaximumSize(screen.width()*3/4, screen.height()*3/4)
        self.info_width = screen.width()*3/4
        self.move(screen.width()/8 , screen.height()/16)

    def create_rcif(self):
        if "f_dir_data" in self._p_d_setup.keys():
            f_dir_data = self._p_d_setup["f_dir_data"]
            if not(os.path.isdir(f_dir_data)):
                f_dir_data = self._p_f_dir_prog
        else:
            f_dir_data = self._p_f_dir_prog
        #f_file_data_new = QtWidgets.QFileDialog.getExistingDirectory(self, "Select a folder:", f_dir_data)
        f_file_data_new, okPressed = QtWidgets.QFileDialog.getSaveFileName(self, "Select a folder:", "full.rcif")
        if not okPressed:
            return
        if f_file_data_new != "":
            f_dir_data = os.path.dirname(f_file_data_new)
            f_name_data = os.path.basename(f_file_data_new)
            self._p_d_setup["f_dir_data"] = f_dir_data
            self._p_d_setup["f_name_data"] = f_name_data
            self.write_setup()
            f_name_full = os.path.join(f_dir_data, f_name_data)

            i, okPressed = QtWidgets.QInputDialog.getInt(self, "Type of data","Type of experiment:\n\n 1. Single crystal\n 2. Powder 1D\n 3 .Powder 2D\n 4 .Powder 2D (one-axial texture)", 1, 1, 4, 1)
            if okPressed:
                exp_type = str(i)
                create_temporary(f_name_full, exp_type)

    def open_rcif(self):
        self.open_file(s_constr="Rcif files (*.rcif)")
        flag_out = self.ask_file(ext=".rcif")
        if not flag_out:
            return None
        self.try_to_plot_from_rcif()


    def open_file(self, s_constr="All files (*.*)"):
        #dir_ = QtWidgets.QFileDialog.getOpenFileName(self, 'Select a folder:', 'C:\\', "XML files (*.xml);;HTML files (*.html);;"
        #        "SVG files (*.svg);;User Interface files (*.ui)")
        if "f_dir_data" in self._p_d_setup.keys():
            f_dir_data = self._p_d_setup["f_dir_data"]
            if not(os.path.isdir(f_dir_data)):
                f_dir_data = self._p_f_dir_prog
        else:
            f_dir_data = self._p_f_dir_prog
        f_file_data_new, okPressed = QtWidgets.QFileDialog.getOpenFileName(self, 'Select a file:', f_dir_data, s_constr)
        if not okPressed:
            return False
        if f_file_data_new != "":
            f_dir_data = os.path.dirname(f_file_data_new)
            f_name_data = os.path.basename(f_file_data_new)
            self._p_d_setup["f_dir_data"] = f_dir_data
            self._p_d_setup["f_name_data"] = f_name_data
            self.write_setup()
        return True
    def ask_file(self, ext=".rcif"):
        cond_ask = False
        cond_cycle = True

        l_key = self._p_d_setup.keys()
        if (("f_name_data" in l_key) & ("f_dir_data" in l_key)):
            f_name_data = self._p_d_setup["f_name_data"]
            f_dir_data = self._p_d_setup["f_dir_data"]
            f_name_full = os.path.join(f_dir_data, f_name_data)
            cond_1 = f_name_data.endswith(ext)
            cond_2 = os.path.isfile(f_name_full)
            cond_ask = not(cond_1 & cond_2)
        else:
            cond_ask = True
        if cond_ask:
            cond_cycle = self.open_file()
        return cond_cycle

    def run_rhochi(self):
        flag_out = self.ask_file(".rcif")
        if not flag_out:
            return None
        f_name_data = self._p_d_setup["f_name_data"]
        f_dir_data = self._p_d_setup["f_dir_data"]
        f_name_full = os.path.join(f_dir_data, f_name_data)
        f_name_out = os.path.join(f_dir_data, "out.rcif")
        
        #f_dir_prog = self._p_f_dir_prog
        #f_prog_full = os.path.join(f_dir_prog, "rhochi.py")
        #line = "python {:} {:} {:}".format(f_prog_full, f_name_full, f_name_full)
        f_dir = os.path.dirname(f_name_full)
        print("Step 1: Input files in directory: ", f_dir)
        rhochi = RhoChi(file_input=f_name_full)
        print("Step 2: File reading")
        rhochi.read_files(f_dir=f_dir)
        print("Step 3: Refienment")
        rhochi.refine()
        print("Step 4: Saving to files")
        rhochi.save_to_files()
        #os.system(line)
        self.try_to_plot_from_rcif()

    def read_rcif(self):
        flag_out = self.ask_file(".rcif")
        if not flag_out:
            return None
        f_name_data = self._p_d_setup["f_name_data"]
        f_dir_data = self._p_d_setup["f_dir_data"]
        f_name_full = os.path.join(f_dir_data, f_name_data)
        os.startfile(f_name_full)

    def plot_data_mono(self, x, y_exp, y_sig):
        widg_central = self._p_widg_central
        stack_widg = widg_central._p_widg_graph
        index = 0
        stack_widg.setCurrentIndex(index)
        widg_graph = stack_widg.widget(index)
        widg_graph.plot_file(x, y_exp, y_sig)

    def plot_data_powder_1d(self, x, l_y_mod, l_y_exp, l_y_sig):
        widg_central = self._p_widg_central
        stack_widg = widg_central._p_widg_graph
        index = 1
        stack_widg.setCurrentIndex(index)
        widg_graph = stack_widg.widget(index)
        widg_graph.plot_file(x, l_y_mod, l_y_exp, l_y_sig)

    def plot_data_powder_2d(self, f_name_full):
        widg_central = self._p_widg_central
        stack_widg = widg_central._p_widg_graph
        index = 2
        stack_widg.setCurrentIndex(index)
        widg_graph = stack_widg.widget(index)
        widg_graph.plot_file(f_name_full)

class cwidget(QtWidgets.QWidget):
    def __init__(self, width):
        self.info_width = width
        super(cwidget, self).__init__()
        self.init_widget()
        self.setLayout(self.layout_central)

    def init_widget(self):
        """
        make central layout
        """
        #lay_central = self.layout()
        #if lay_central == None:
        #    lay_central = QtWidgets.QHBoxLayout()
        #else:
        #    del_layout(lay_central)
        lay_central = QtWidgets.QHBoxLayout()


        width_m_1 = self.info_width/10
        width_m_2 = (4*self.info_width)/10
        width_m_3 = self.info_width-width_m_1-width_m_2-20

        self.width_cpanel = width_m_1
        self.width_left = width_m_2
        self.width_right = width_m_3


        self.lay_cpanel = QtWidgets.QVBoxLayout()
        self.lay_left = QtWidgets.QVBoxLayout()
        self.lay_right = QtWidgets.QVBoxLayout()


        lay_central.addLayout(self.lay_cpanel)
        lay_central.addLayout(self.lay_left)
        lay_central.addLayout(self.lay_right)

        self.layout_central = lay_central

        self.form_widg_cpanel()
        self.lay_left.addWidget(QtWidgets.QFrame())
        self.form_widg_right()
        


    def form_widg_cpanel(self):


        #del_layout(self.lay_cpanel)

        width_m_1 = self.width_cpanel


        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)



    def form_widg_right(self):

        width_m_3 = self.width_right
        #widg_powder_1d = QtWidgets.QFrame()
        stack_widg = QtWidgets.QStackedWidget(self)
        stack_widg.setMinimumSize(width_m_3, 1)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        stack_widg.setSizePolicy(sizePolicy)


        widg_mono = interactive_graph_mod_mono.cwidg_central()
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        widg_mono.setSizePolicy(sizePolicy)
        stack_widg.addWidget(widg_mono)
        
        widg_powder_1d = interactive_graph_mod_pwd.cwidg_central()
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        widg_powder_1d.setSizePolicy(sizePolicy)

        stack_widg.addWidget(widg_powder_1d)

        widg_powder_2d = interactive_graph_mod_pwd_2d_as_1d.cwidg_central()
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        widg_powder_2d.setSizePolicy(sizePolicy)

        stack_widg.addWidget(widg_powder_2d)
        
        self.lay_right.addWidget(stack_widg)
        self._p_widg_graph = stack_widg
        stack_widg.setCurrentIndex(0)

def main(l_arg= []):
    app = QtWidgets.QApplication(l_arg)
    f_dir_prog = os.path.dirname(__file__)
    mainwind1 = cbuilder(f_dir_prog)
    sys.exit(app.exec_())


if __name__ == '__main__':
    l_arg = sys.argv
    main(l_arg)





"""
define classes to describe AtomSiteMagnetism
"""
__author__ = 'ikibalin'
__version__ = "2019_08_29"
import os
import numpy


from pycifstar import Global
from cryspy.f_common.cl_fitable import Fitable

class AtomSiteMagnetism(object):
    """
    Data items in the ATOM_SITE_MAGNETISM category record details about
    magnetic properties of the atoms that occupy the atom sites.
    
    Description in cif file:

    loop_                                     
    _atom_site_magnetism_label
    _atom_site_magnetism_lande
    _atom_site_magnetism_kappa
    Fe3A 2.0 1.0
    Fe3B 2.0 1.0

    
    """    
    def __init__(self, label=[], lande=[], kappa=[]):
        super(AtomSiteMagnetism, self).__init__()

        self.__atom_site_magnetism_label = []
        self.__atom_site_magnetism_lande = []
        self.__atom_site_magnetism_kappa = []
        self.label = label
        self.lande = lande
        self.kappa = kappa
        
    def __repr__(self):
        ls_out = ["AtomSiteMagnetism:"]
        ls_out.append(" label    kappa    lande   ")
        for hh_1, hh_2, hh_3 in zip(self.label, self.kappa, self.lande):
            ls_out.append(" {:8} {:8} {:8}".format(hh_1, hh_2.print_with_sigma, hh_3.print_with_sigma))
        return "\n".join(ls_out)

    @property
    def label(self):
        """
        The _atom_site_magnetism_label is a unique identifier for a particular site
        in the crystal. 

        Type: char
        """
        return tuple(self.__atom_site_magnetism_label)
    @label.setter
    def label(self, l_x):
        l_x_in = []
        for x in l_x:
            x_in = str(x).strip()
            l_x_in.append(x_in)
        self.__atom_site_magnetism_label = l_x_in
        len_x = len(l_x_in)
        len_1 = len(self.__atom_site_magnetism_lande)
        if len_1 > len_x:
            self.__atom_site_magnetism_lande = self.__atom_site_magnetism_lande[:len_x]
        elif len_1 < len_x:
            l_fitable = [Fitable(value=2., name="_atom_site_magnetism_lande") for hh in range(len_x-len_1)]
            self.__atom_site_magnetism_lande.extend(l_fitable)

        len_1 = len(self.__atom_site_magnetism_kappa)
        if len_1 > len_x:
            self.__atom_site_magnetism_kappa = self.__atom_site_magnetism_kappa[:len_x]
        elif len_1 < len_x:
            l_fitable = [Fitable(value=1., name="_atom_site_magnetism_kappa") for hh in range(len_x-len_1)]
            self.__atom_site_magnetism_kappa.extend(l_fitable)

    @property
    def lande(self):
        """
        Lande factor. 

        Default: 2

        Type: float
        """
        return tuple(self.__atom_site_magnetism_lande)
    @lande.setter
    def lande(self, l_x):
        l_fitable = []
        for x in l_x:
            if isinstance(x, Fitable):
                x_in = x
            else:
                x_in = Fitable()
                flag = x_in.take_it(x)
            l_fitable.append(x_in)
        len_x = len(l_fitable)
        len_1 = len(self.__atom_site_magnetism_label)
        if len_1 < len_x:
            l_fitable = l_fitable[:len_1]
        elif len_1 > len_x:
            l_fitable.extend([Fitable(value=2., name="_atom_site_magnetism_lande") for hh in range(len_1-len_x)])
        self.__atom_site_magnetism_lande = l_fitable

    @property
    def kappa(self):
        """
        kappa 

        Default: 1

        Type: float
        """
        return tuple(self.__atom_site_magnetism_kappa)
    @kappa.setter
    def kappa(self, l_x):
        l_fitable = []
        for x in l_x:
            if isinstance(x, Fitable):
                x_in = x
            else:
                x_in = Fitable()
                flag = x_in.take_it(x)
            l_fitable.append(x_in)
        len_x = len(l_fitable)
        len_1 = len(self.__atom_site_magnetism_label)
        if len_1 < len_x:
            l_fitable = l_fitable[:len_1]
        elif len_1 > len_x:
            l_fitable.extend([Fitable(value=1., name="_atom_site_magnetism_kappa") for hh in range(len_1-len_x)])
        self.__atom_site_magnetism_kappa = l_fitable

    def _show_message(self, s_out: str):
        print("***  Error ***")
        print(s_out)

    @property
    def is_defined(self):
        """
        Output: True if all started parameters are given
        """
        cond = (self.label != [])
        return cond
    @property
    def is_variable(self):
        res = (any([hh.refinement for hh in self.lande]) | any([hh.refinement for hh in self.kappa]))
        return res


    def get_variables(self):
        l_variable = [hh for hh in self.lande if hh.refinement]
        l_variable.extend([hh for hh in self.kappa if hh.refinement])

        return l_variable

    @property
    def to_cif(self):
        ls_out = []
        if self.is_defined:
            ls_out.append("loop_")
            ls_out.append("_atom_site_magnetism_label")
            ls_out.append("_atom_site_magnetism_lande")
            ls_out.append("_atom_site_magnetism_kappa")
            for hh_1, hh_2, hh_3 in zip(self.label, self.lande, self.kappa):
                ls_out.append("{:} {:} {:}".format(hh_1, 
                                                   hh_2.print_with_sigma, 
                                                   hh_3.print_with_sigma))
        return "\n".join(ls_out)

    def from_cif(self, string: str):
        cif_global = Global()
        flag = cif_global.take_from_string(string)
        if not flag:
            return False
        flag = False
        flag = cif_global.is_prefix("_atom_site_magnetism")
        if flag:
            cif_loop = cif_global["_atom_site_magnetism"]
            l_name = cif_loop.names
            if "_atom_site_magnetism_label" in l_name:
                self.label = cif_loop["_atom_site_magnetism_label"]

            if "_atom_site_magnetism_lande" in l_name:
                l_val = cif_loop["_atom_site_magnetism_lande"]
                l_fitable = []
                for val in l_val:
                    fitable = Fitable(name="_atom_site_magnetism_lande")
                    fitable.take_it(val)
                    l_fitable.append(fitable)
                self.lande = l_fitable
            if "_atom_site_magnetism_kappa" in l_name:
                l_val = cif_loop["_atom_site_magnetism_kappa"]
                l_fitable = []
                for val in l_val:
                    fitable = Fitable(name="_atom_site_magnetism_kappa")
                    fitable.take_it(val)
                    l_fitable.append(fitable)
                self.kappa = l_fitable
        else:
            self.label = []

        return True

    def _form_lande_kappa(self, atom_site):
        label = numpy.array(atom_site.label, dtype=str)
        label_magnetism = numpy.array(self.label, dtype=str)
        if not(set(label_magnetism).issubset(set(label))):
            self._show_message("Unknown 'aniso_label'")
            return False

        np_index = numpy.array([int(numpy.argwhere(label==hh)[0]) for hh in label_magnetism], dtype=int)

        lande_in = 2.*numpy.ones(label.shape, dtype=float)
        kappa_in = numpy.ones(label.shape, dtype=float)

        lande = numpy.array(self.lande, dtype=float)
        kappa = numpy.array(self.kappa, dtype=float)

        lande_in[np_index], kappa_in[np_index] = lande, kappa

        return lande_in, kappa_in


from .cl_pd_background import PdBackground
from .cl_pd_exclude_2theta import PdExclude2Theta
from .cl_pd_instr_reflex_asymmetry import PdInstrReflexAsymmetry
from .cl_pd_instr_resolution import PdInstrResolution
from .cl_pd_meas import PdMeas
from .cl_pd_peak import PdPeak
from .cl_pd_phase import PdPhase
from .cl_pd_proc import PdProc
from .cl_pd import Pd



from .cl_beam_polarization import BeamPolarization
from .cl_orient_matrix import OrientMatrix
from .cl_extinction import Extinction


from .f_single import (
    DiffrnRefln, 
    Diffrn
    )


from .f_powder_1d import (
    PdBackground,
    PdExclude2Theta,
    PdInstrReflexAsymmetry,
    PdInstrResolution,
    PdMeas,
    PdPeak,
    PdPhase,
    PdProc,
    Pd
    )

from .f_powder_2d import (
    Pd2dBackground,
    Pd2dExclude,
    Pd2dInstrReflexAsymmetry,
    Pd2dInstrResolution,
    Pd2dMeas,
    Pd2dPhase,
    Pd2dPeak,
    Pd2dProc,
    Pd2d
    )

from .f_powder_texture_2d import (
    Pd2dt
    )



"""
define classe Pd2dExclude which describes the excluded regions in 2d powder diffraction experiment
"""
__author__ = 'ikibalin'
__version__ = "2019_09_10"
import os
import numpy


from pycifstar import Global


class Pd2dExclude(object):
    """
    Example:

    loop_
    _pd2d_exclude_phi_min
    _pd2d_exclude_phi_max
    _pd2d_exclude_tth_min
    _pd2d_exclude_tth_max
     -10.0 60.0 4.0 10.0 
    """
    def __init__(self, ttheta_min=[], ttheta_max=[], phi_min=[], phi_max=[]
                 ):
        super(Pd2dExclude, self).__init__()
        self.__pd2d_exclude_2theta_min = None
        self.__pd2d_exclude_2theta_max = None
        self.__pd2d_exclude_phi_min = None
        self.__pd2d_exclude_phi_max = None

        self.ttheta_min = ttheta_min
        self.ttheta_max = ttheta_max
        self.phi_min = phi_min
        self.phi_max = phi_max

    @property
    def ttheta_min(self):
        return self.__pd2d_exclude_2theta_min
    @ttheta_min.setter
    def ttheta_min(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, float):
                x_in = x
            else:
                x_in = float(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=float)
        self.__pd2d_exclude_2theta_min = np_x_in

    @property
    def ttheta_max(self):
        return self.__pd2d_exclude_2theta_max
    @ttheta_max.setter
    def ttheta_max(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, float):
                x_in = x
            else:
                x_in = float(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=float)
        self.__pd2d_exclude_2theta_max = np_x_in


    @property
    def phi_min(self):
        return self.__pd2d_exclude_phi_min
    @phi_min.setter
    def phi_min(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, float):
                x_in = x
            else:
                x_in = float(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=float)
        self.__pd2d_exclude_phi_min = np_x_in

    @property
    def phi_max(self):
        return self.__pd2d_exclude_phi_max
    @phi_max.setter
    def phi_max(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, float):
                x_in = x
            else:
                x_in = float(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=float)
        self.__pd2d_exclude_phi_max = np_x_in



    def __repr__(self):
        ls_out = ["Pd2dExclude:"]
        ls_out.append("ttheta_min ttheta_max    phi_min    phi_max")
        for _1, _2, _3, _4 in zip(self.ttheta_min, self.ttheta_max, self.phi_min, self.phi_max):
            ls_out.append("{:10.2f} {:10.2f} {:10.2f} {:10.2f}".format(_1, _2, _3, _4))
        return "\n".join(ls_out)

    @property
    def to_cif(self):
        ls_out = []
        if self.is_defined:
            ls_out.append("loop_")
            ls_out.append("_pd2d_exclude_2theta_min")
            ls_out.append("_pd2d_exclude_2theta_max")
            ls_out.append("_pd2d_exclude_phi_min")
            ls_out.append("_pd2d_exclude_phi_max")
            for _1, _2, _3, _4 in zip(self.ttheta_min, self.ttheta_max, self.phi_min, self.phi_max):
                ls_out.append("{:} {:} {:} {:}".format(_1, _2, _3, _4))
        return "\n".join(ls_out)

    def from_cif(self, string: str):
        cif_global = Global()
        flag = cif_global.take_from_string(string)
        if not flag:
            return False
        flag = False
        flag = cif_global.is_prefix("_pd2d_exclude")
        if flag:
            cif_loop = cif_global["_pd2d_exclude"]
            l_name = cif_loop.names
            if "_pd2d_exclude_2theta_min" in l_name:
                self.ttheta_min = [float(_1) for _1 in cif_loop["_pd2d_exclude_2theta_min"]]
            if "_pd2d_exclude_2theta_max" in l_name:
                self.ttheta_max = [float(_1) for _1 in cif_loop["_pd2d_exclude_2theta_max"]]
            if "_pd2d_exclude_phi_min" in l_name:
                self.phi_min = [float(_1) for _1 in cif_loop["_pd2d_exclude_phi_min"]]
            if "_pd2d_exclude_phi_max" in l_name:
                self.phi_max = [float(_1) for _1 in cif_loop["_pd2d_exclude_phi_max"]]
        else:
            self.ttheta_min, self.ttheta_max, self.phi_min, self.phi_max = [], [], [], []
        return True

    @property
    def is_defined(self):
        cond = all([self.ttheta_min is not None, self.ttheta_max is not None, self.phi_min is not None, self.phi_max is not None])
        return cond

    @property
    def is_variable(self):
        return False
    
    def get_variables(self):
        return []

    def _show_message(self, s_out: str):
        print("***  Error ***")
        print(s_out)



"""
RhoChi program
"""
__author__ = 'ikibalin'
__version__ = "2019_09_15"
import os
import sys
import numpy
import scipy.optimize
import time

from pycifstar import Global
from cryspy.f_common.cl_fitable import Fitable

from cryspy.f_crystal.cl_cell import Cell
from cryspy.f_crystal.cl_space_group import SpaceGroup
from cryspy.f_crystal.cl_atom_type import AtomType
from cryspy.f_crystal.cl_atom_site import AtomSite
from cryspy.f_crystal.cl_atom_site_aniso import AtomSiteAniso
from cryspy.f_crystal.cl_atom_site_magnetism import AtomSiteMagnetism
from cryspy.f_crystal.cl_atom_site_magnetism_aniso import AtomSiteMagnetismAniso

from cryspy.f_experiment.f_single.cl_diffrn_refln import DiffrnRefln
from cryspy.f_experiment.f_powder_1d.cl_pd_meas import PdMeas
from cryspy.f_experiment.f_powder_1d.cl_pd_phase import PdPhase
from cryspy.f_experiment.f_powder_1d.cl_pd_background import PdBackground
from cryspy.f_experiment.f_powder_2d.cl_pd2d_meas import Pd2dMeas
from cryspy.f_experiment.f_powder_2d.cl_pd2d_phase import Pd2dPhase
from cryspy.f_experiment.f_powder_2d.cl_pd2d_background import Pd2dBackground

from cryspy.f_crystal.cl_crystal import Crystal
from cryspy.f_experiment.f_single.cl_diffrn import Diffrn
from cryspy.f_experiment.f_powder_1d.cl_pd import Pd
from cryspy.f_experiment.f_powder_2d.cl_pd2d import Pd2d
from cryspy.f_experiment.f_powder_texture_2d.cl_pd2dt import Pd2dt




class RhoChi(dict):
    """
    Class to describe rhochi container
    """
    def __init__(self, label="", file_out=None, file_dir=None, experiments=[], crystals=[], file_input=None):
        super(RhoChi, self).__init__()
        self.__label = None
        self.__experiments = []
        self.__crystals = []
        self.__file_input = None
        self.__file_out = None
        self.__file_dir = None

        self.label = label
        self.crystals = crystals
        self.experiments = experiments
        self.file_dir = file_dir
        self.file_out = file_out
        self.file_input = file_input

    @property
    def label(self):
        return self.__label
    @label.setter
    def label(self, x: str):
        self.__label = x

    @property
    def experiments(self):
        return self.__experiments
    @experiments.setter
    def experiments(self, l_x):
        self.__experiments = l_x

    @property
    def crystals(self):
        return self.__crystals
    @crystals.setter
    def crystals(self, l_x):
        self.__crystals = l_x

    @property
    def file_input(self):
        return self.__file_input
    @file_input.setter
    def file_input(self, x: str):
        self.__file_input = x

    @property
    def file_out(self):
        return self.__file_out
    @file_out.setter
    def file_out(self, x: str):
        self.__file_out = x

    @property
    def file_dir(self):
        return self.__file_dir
    @file_dir.setter
    def file_dir(self, x: str):
        self.__file_dir = x

    def __repr__(self):
        ls_out = []
        ls_out.append("RhoChi:")
        ls_out.append(" label: ".format(self.label))
        ls_out.append(" file_out: ".format(self.file_out))
        ls_out.append("\n Crystals:")
        ls_out.append("\n".join([str(_) for _ in self.crystals]))
        ls_out.append("\n Experiments:")
        ls_out.append("\n".join([str(_) for _ in self.experiments]))
        return "\n".join(ls_out)

    
    def calc_chi_sq(self):
        """
        calculate the integral intensity for h, k, l reflections
        """
        self.apply_constraint()
            
        l_crystal = self.crystals

        chi_sq_res, n_res = 0., 0.
        for experiment in self.experiments:
            chi_sq, n = experiment.calc_chi_sq(l_crystal)
            experiment.chi_sq = chi_sq
            experiment.n = n
            chi_sq_res += chi_sq
            n_res += n
        return chi_sq_res, n_res
    
    def apply_constraint(self):
        for crystal in self.crystals:
            crystal.apply_constraint()

    def _show_message(self, s_out: str):
        print("***  Error ***")
        print(s_out)

    def refine(self):
        """
        optimization
        """
        self.apply_constraint()
        l_fitable = self.get_variables()

        if l_fitable == []:
            self._show_message("Variables are not found")
            chi_sq, n = self.calc_chi_sq()
            self._show_message("Chi_sq {:.3f}\npoints: {:}".format(chi_sq, n))
            return None
        

        val_0 = numpy.array([fitable.value for fitable in l_fitable], dtype=float)
        
        sign = 2*(numpy.array(val_0 >= 0., dtype=int)-0.5)
        param_0 = numpy.log(abs(val_0)*(numpy.e-1.)+1.)*sign
        coeff_norm = numpy.where(val_0 == 0., 1., val_0)/numpy.where(param_0==0., 1., param_0)
        hes_coeff_norm = numpy.matmul(coeff_norm[:, numpy.newaxis], coeff_norm[numpy.newaxis, :])

        def tempfunc(l_param):
            for fitable, param, _1 in zip(l_fitable, l_param, coeff_norm):
                fitable.value = param*_1
            chi_sq, n_points = self.calc_chi_sq()
            return (chi_sq*1./float(n_points))

        chi_sq, n = self.calc_chi_sq()

        #if self.ref[0].refin:
        print("starting chi_sq/n: {:.2f} (n = {:}).".format(chi_sq*1./n, int(n)))
        print("\nrefinement started for parameters:")
        ls_out = " ".join(["{:12}".format(fitable.name.rjust(12)) if len(fitable.name)<=12 
                           else "{:12}".format(fitable.name[-12:]) for fitable in l_fitable]) + "       chi_sq"
        print(ls_out)
        aa = time.time()
        """
        res, m_error, infodict, errmsg, ier = \
            scipy.optimize.leastsq(tempfunc, param_0, full_output=1)

        """
        res = scipy.optimize.minimize(tempfunc, param_0, method='BFGS', callback=lambda x : self._f_callback(coeff_norm, x), options = {"disp": True})

        #res = scipy.optimize.minimize(tempfunc, l_param_0, method='Nelder-Mead', 
        #                              callback=self._f_callback, options = {"fatol": 0.01*n})

        bb = time.time()
        print("refinement complete, time {:.2f} sec.\n\nfinal chi_sq/n: {:.2f}\nstarted chi_sq/n: {:.2f}".format(bb-aa, res.fun, chi_sq*1./n))
        
        hess_inv = res["hess_inv"] * hes_coeff_norm
        sigma = (abs(numpy.diag(hess_inv)*1./float(n)))**0.5
        for fitable, _1  in zip(l_fitable, sigma):
            fitable.sigma = _1
        print("experiment  chi_sq_n")
        for _1 in self.experiments:
            print("{:10}: {:8.2f}".format(_1.label, _1.chi_sq/_1.n))
        return res

    
    def _f_callback(self, *arg):
        coeff_norm = arg[0]
        res_x = arg[1]
        ls_out = " ".join(["{:12.5f}".format(_1*_2) for _1, _2 in zip(res_x, coeff_norm)])
        if len(arg) > 2:
            res_fun = arg[1]
            ls_out += " {:12.1f}".format(res_fun.fun)
        print(ls_out)
    
    @property
    def is_variable(self):
        res = (any([_.is_variable for _ in self.experiments]) |
               any([_.is_variable for _ in self.crystals]))
        return res

    def get_variables(self):
        l_variable = []
        for _ in self.crystals:
            l_variable.extend(_.get_variables())
        for _ in self.experiments:
            l_variable.extend(_.get_variables())
        return l_variable

    @property
    def to_cif(self):
        ls_out = []
        ls_out.append("global_{:}\n".format(self.label))
        if self.crystals is not None:
            ls_out.extend([_.to_cif for _ in self.crystals])
        if self.experiments is not None:
            ls_out.extend([_.to_cif for _ in self.experiments])
        return "\n".join(ls_out)

    def from_cif(self, string: str):
        l_crystal, l_experiment = [], []
        cif_global = Global()
        flag = cif_global.take_from_string(string)
        if not flag:
            return False
        self.label = cif_global.name
        l_data = cif_global.datas
        for data in l_data:
            flag_crystal, flag_diffrn, flag_pd, flag_pd2d = False, False, False, False
            flag_crystal = data.is_prefix("_cell_length_a")
            flag_diffrn = data.is_prefix("_diffrn_refln") | data.is_prefix("_diffrn_orient_matrix_UB")
            flag_pd = data.is_prefix("_pd_meas") | data.is_prefix("_pd_instr_resolution")
            flag_pd2d = data.is_prefix("_pd2d_meas") | data.is_prefix("_pd2d_instr_resolution")
            flag_pd2dt = data.is_prefix("_2dpdt_texture")
            if flag_crystal:
                crystal = Crystal()
                crystal.from_cif(str(data))
                l_crystal.append(crystal)

            if flag_diffrn:
                diffrn = Diffrn()
                diffrn.from_cif(str(data))
                l_experiment.append(diffrn)

            if flag_pd:
                pd = Pd()
                pd.from_cif(str(data))
                l_experiment.append(pd)

            if (flag_pd2d & (not(flag_pd2dt))):
                pd2d = Pd2d()
                pd2d.from_cif(str(data))
                l_experiment.append(pd2d)
            elif (flag_pd2d & flag_pd2dt):
                pd2dt = Pd2dt()
                pd2dt.from_cif(str(data))
                l_experiment.append(pd2dt)

        self.experiments = l_experiment
        self.crystals = l_crystal

    def read_file(self, f_name):
        self.file_input = f_name
        with open(f_name, "r") as fid:
            string = fid.read()
            self.from_cif(string)
        self.apply_constraint()

    def save_to_file(self, f_name):
        self.file_input = f_name
        if os.path.basename(f_name) == "main.rcif":
            self.save_to_files()
        else:
            with open(f_name, "w") as fid:
                fid.write(self.to_cif)


    def save_to_files(self):
        if self.file_input is None:
            f_dir = "."
        else:
            f_dir = os.path.dirname(self.file_input)
        f_main = os.path.join(f_dir, "main.rcif")
        ls_main = []
        for crystal in self.crystals:
            ls_main.append("\n"+crystal.to_cif)
        for experiment in self.experiments:   
            ls_main.append("\ndata_{:}".format(experiment.label))
            ls_main.append(experiment.params_to_cif)

            f_data = os.path.join(f_dir, "{:}_data.rcif".format(experiment.label))
            ls_data = []
            ls_data.append("\ndata_{:}".format(experiment.label))
            ls_data.append(experiment.data_to_cif)
            with open(f_data, 'w') as fid:
                fid.write("\n".join(ls_data))
            f_calc = os.path.join(f_dir, "{:}_calc.rcif".format(experiment.label))
            ls_calc = []
            ls_calc.append("\ndata_{:}".format(experiment.label))
            ls_calc.append(experiment.calc_to_cif)
            with open(f_calc, 'w') as fid:
                fid.write("\n".join(ls_calc))
        with open(f_main, 'w') as fid:
            fid.write("\n".join(ls_main))

    def read_files(self, f_dir="."):
        f_main = os.path.join(f_dir, "main.rcif")
        self.file_input = f_main
        
        if not(os.path.isfile(f_main)):
            self._show_message("File '{:}' is not found.".format(f_main))
            return None
        with open(f_main, "r") as fid:
            string = fid.read()
        self.from_cif(string)
        
        for experiment in self.experiments:
            label_orig = experiment.label
            f_data = os.path.join(f_dir, "{:}_data.rcif".format(label_orig))
            if not(os.path.isfile(f_main)):
                self._show_message("File '{:}' is not found.".format(f_data))
            else:
                with open(f_data, "r") as fid:
                    string = fid.read()
                experiment.from_cif(string)
                experiment.label = label_orig
        self.apply_constraint()


def rhochi_read_file(f_name):
    rho_chi = RhoChi(file_input=f_name)
    if os.path.basename(f_name) == "main.rcif":
        rho_chi.read_files(os.path.dirname(f_name))
    else:
        rho_chi.read_file(f_name)
    return rho_chi

def rhochi_read_files():
    rho_chi = RhoChi()
    rho_chi.read_files(".")
    return rho_chi

def rhochi_refinement(f_name_in="", f_name_out=""):
    """
    refinement,
    parameters are defined in given .rcif fiel
    """
    print(70*"*"+"\n"+"RhoChi program. Console version.".center(70)+"\n"+
          70*"*")
    if f_name_in != "":
        rho_chi = rhochi_read_file(f_name_in)
    else:
        rho_chi = rhochi_read_files()
    print("Before refinement:\n")
    #print(rho_chi)
    print("\nRefined parameters -- before:\n")
    for fitable in rho_chi.get_variables():
        print(fitable)

    rho_chi.refine()

    print("After refinement:\n")
    #print(rho_chi)
    print("\nRefined parameters -- after:\n")
    for fitable in rho_chi.get_variables():
        print(fitable)

    if f_name_out != "":
        rho_chi.save_to_file(f_name_out)
    else:
        rho_chi.save_to_files()

    print(70*"*"+"\n"+70*"*")


def create_temporary(f_name_in, exp_type="2"):
    atom_site = AtomSite(label=["Fe"], type_symbol=["Fe3+"], x=[Fitable(0.125)],
                         y=[Fitable(0.125)], z=[Fitable(0.125)], adp_type=["uani"],
                         b_iso=[Fitable(0.0)], occupancy=[Fitable(0.)])
    space_group = SpaceGroup(spgr_given_name="Fd-3m", spgr_choice="2")
    cell = Cell(a=Fitable(8.3), bravais_lattice="cubic")
    atom_site_aniso = AtomSiteAniso(label=["Fe"], u_11=[Fitable(0.)], u_22=[Fitable(0.)],
                                    u_33=[Fitable(0.)], u_12=[Fitable(0.)], u_13=[Fitable(0.)], 
                                    u_23=[Fitable(0.)])
    atom_site_magnetism = AtomSiteMagnetism(label=["Fe"], lande=[Fitable(2.)], kappa=[Fitable(1.)])
    atom_site_magnetism_aniso = AtomSiteMagnetismAniso(label=["Fe"], chi_type=["cani"], 
                                    chi_11=[Fitable(0.)], chi_22=[Fitable(0.)], chi_33=[Fitable(0.)], 
                                    chi_12=[Fitable(0.)], chi_13=[Fitable(0.)], chi_23=[Fitable(0.)])

    crystal = Crystal(label="phase1", cell=cell, space_group=space_group, atom_site=atom_site,
                      atom_site_aniso=atom_site_aniso, atom_site_magnetism=atom_site_magnetism,
                      atom_site_magnetism_aniso=atom_site_magnetism_aniso)
    
    if "1" in exp_type:
        diffrn_refln = DiffrnRefln(h=[1], k=[0], l=[0], fr=[1], fr_sigma=[0.1])
        experiment = Diffrn(diffrn_refln=diffrn_refln)
    if "2" in exp_type:
        pd_phase = PdPhase(label=["phase1"], scale=[Fitable(1.)], igsize=[Fitable(0.)])
        pd_meas = PdMeas(ttheta=[4., 25.], up=[100., 150.], up_sigma=[1., 1.], down=[80., 170.],
                    down_sigma=[8., 10.])
        pd_background = PdBackground(ttheta=[4.5, 80], intensity=[0., 0.])
        experiment = Pd(meas=pd_meas, phase=pd_phase, background=pd_background)
    if (("3" in exp_type) | ("4" in exp_type)):
        pd2d_phase = Pd2dPhase(label=["phase1"], scale=[Fitable(1.)], igsize=[Fitable(0.)])
        pd2d_meas = Pd2dMeas(ttheta=[3.,24.],phi=[0,5], 
                            up=[[10,12], [11,17]], up_sigma=[[1,1], [1,2]], 
                            down=[[11,11], [12,14]], down_sigma=[[1,1], [1,2]])
        pd2d_background = Pd2dBackground(ttheta=[4.5, 80], phi=[0., 80.], intensity=[[0., 0.], [0., 0.]])
        if "3" in exp_type:   
            experiment = Pd2d(meas=pd2d_meas, phase=pd2d_phase, background=pd2d_background)
        else: #elif "4"   
            experiment = Pd2dt(meas=pd2d_meas, phase=pd2d_phase, background=pd2d_background)

    rhochi = RhoChi(crystals=[crystal], experiments = [experiment], file_input=f_name_in)
    rhochi.save_to_files()




"""
define class Refln
"""
__author__ = 'ikibalin'
__version__ = "2019_09_09"
import os
import numpy


from pycifstar import Global


class Refln(object):
    """
    Data items in the REFLN category record details about the
    reflections used to determine the ATOM_SITE data items.

    The REFLN data items refer to individual reflections and must
    be included in looped lists.

    The REFLNS data items specify the parameters that apply to all
    reflections. The REFLNS data items are not looped.


    Example:

    loop_
     _refln_index_h
     _refln_index_k
     _refln_index_l
     _refln_d_spacing
     _refln_A_calc
     _refln_B_calc
     _refln_chi_11_A_calc
     _refln_chi_12_A_calc
     _refln_chi_13_A_calc
     _refln_chi_21_A_calc
     _refln_chi_22_A_calc
     _refln_chi_23_A_calc
     _refln_chi_31_A_calc
     _refln_chi_32_A_calc
     _refln_chi_33_A_calc
     _refln_chi_11_B_calc
     _refln_chi_12_B_calc
     _refln_chi_13_B_calc
     _refln_chi_21_B_calc
     _refln_chi_22_B_calc
     _refln_chi_23_B_calc
     _refln_chi_31_B_calc
     _refln_chi_32_B_calc
     _refln_chi_33_B_calc
    
    reference: https://www.iucr.org/__data/iucr/cifdic_html/1/cif_core.dic/Crefln.html
    """
    def __init__(self, h=[], k=[], l=[], f_nucl=[], sft_11=[], sft_12=[], sft_13=[], sft_21=[], sft_22=[], sft_23=[], sft_31=[], sft_32=[], sft_33=[]):
        super(Refln, self).__init__()
        self.__refln_index_h = None
        self.__refln_index_k = None
        self.__refln_index_l = None
        self.__refln_d_spacing = None
        self.__refln_f_nucl_calc = None
        self.__refln_sft_11_calc = None
        self.__refln_sft_12_calc = None
        self.__refln_sft_13_calc = None
        self.__refln_sft_21_calc = None
        self.__refln_sft_22_calc = None
        self.__refln_sft_23_calc = None
        self.__refln_sft_31_calc = None
        self.__refln_sft_32_calc = None
        self.__refln_sft_33_calc = None
        self.h = h
        self.k = k
        self.l = l
        self.f_nucl = f_nucl
        self.sft_11 = sft_11
        self.sft_12 = sft_12
        self.sft_13 = sft_13
        self.sft_21 = sft_21
        self.sft_22 = sft_22
        self.sft_23 = sft_23
        self.sft_31 = sft_31
        self.sft_32 = sft_32
        self.sft_33 = sft_33

    @property
    def h(self):
        return self.__refln_index_h
    @h.setter
    def h(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, int):
                x_in = x
            else:
                x_in = int(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=int)
        self.__refln_index_h = np_x_in

    @property
    def k(self):
        return self.__refln_index_k
    @k.setter
    def k(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, int):
                x_in = x
            else:
                x_in = int(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=int)
        self.__refln_index_k = np_x_in

    @property
    def l(self):
        return self.__refln_index_l
    @l.setter
    def l(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, int):
                x_in = x
            else:
                x_in = int(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=int)
        self.__refln_index_l = np_x_in

    @property
    def f_nucl(self):
        return self.__refln_f_nucl_calc
    @f_nucl.setter
    def f_nucl(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, complex):
                x_in = x
            else:
                x_in = complex(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=complex)
        self.__refln_f_nucl_calc = np_x_in

    @property
    def sft_11(self):
        return self.__refln_sft_11_calc
    @sft_11.setter
    def sft_11(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, complex):
                x_in = x
            else:
                x_in = complex(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=complex)
        self.__refln_sft_11_calc = np_x_in

    @property
    def sft_12(self):
        return self.__refln_sft_12_calc
    @sft_12.setter
    def sft_12(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, complex):
                x_in = x
            else:
                x_in = complex(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=complex)
        self.__refln_sft_12_calc = np_x_in

    @property
    def sft_13(self):
        return self.__refln_sft_13_calc
    @sft_13.setter
    def sft_13(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, complex):
                x_in = x
            else:
                x_in = complex(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=complex)
        self.__refln_sft_13_calc = np_x_in

    @property
    def sft_21(self):
        return self.__refln_sft_21_calc
    @sft_21.setter
    def sft_21(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, complex):
                x_in = x
            else:
                x_in = complex(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=complex)
        self.__refln_sft_21_calc = np_x_in

    @property
    def sft_22(self):
        return self.__refln_sft_22_calc
    @sft_22.setter
    def sft_22(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, complex):
                x_in = x
            else:
                x_in = complex(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=complex)
        self.__refln_sft_22_calc = np_x_in

    @property
    def sft_23(self):
        return self.__refln_sft_23_calc
    @sft_23.setter
    def sft_23(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, complex):
                x_in = x
            else:
                x_in = complex(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=complex)
        self.__refln_sft_23_calc = np_x_in

    @property
    def sft_31(self):
        return self.__refln_sft_31_calc
    @sft_31.setter
    def sft_31(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, complex):
                x_in = x
            else:
                x_in = complex(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=complex)
        self.__refln_sft_31_calc = np_x_in

    @property
    def sft_32(self):
        return self.__refln_sft_32_calc
    @sft_32.setter
    def sft_32(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, complex):
                x_in = x
            else:
                x_in = complex(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=complex)
        self.__refln_sft_32_calc = np_x_in

    @property
    def sft_33(self):
        return self.__refln_sft_33_calc
    @sft_33.setter
    def sft_33(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, complex):
                x_in = x
            else:
                x_in = complex(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=complex)
        self.__refln_sft_33_calc = np_x_in

            
    def __repr__(self):
        ls_out = ["Refln:"]
        ls_out.append("   h   k   l           A           B    sft_11_A    sft_12_A    sft_13_A      sft_11_B    sft_12_B    sft_13_B")
        ls_out.append("                                        sft_21_A    sft_22_A    sft_23_A      sft_21_B    sft_22_B    sft_23_B")
        ls_out.append("                                        sft_31_A    sft_32_A    sft_33_A      sft_31_B    sft_32_B    sft_33_B")
        for _1, _2, _3, _4, _5, _6, _7, _8, _9, _10, _11, _12, _13 in zip(self.h, self.k, self.l, self.f_nucl, 
                                                                          self.sft_11, self.sft_12, self.sft_13, 
                                                                          self.sft_21, self.sft_22, self.sft_23, 
                                                                          self.sft_31, self.sft_32, self.sft_33):
            ls_out.append("\n{:4}{:4}{:4}{:12.5f}{:12.5f}{:12.5f}{:12.5f}{:12.5f}  {:12.5f}{:12.5f}{:12.5f}".format(
                _1, _2, _3, _4.real, _4.imag, _5.real, _6.real, _7.real, _5.imag, _6.imag, _7.imag))
            ls_out.append("                                    {:12.5f}{:12.5f}{:12.5f}  {:12.5f}{:12.5f}{:12.5f}".format(
                _8.real, _9.real, _10.real, _8.imag, _9.imag, _10.imag))
            ls_out.append("                                    {:12.5f}{:12.5f}{:12.5f}  {:12.5f}{:12.5f}{:12.5f}".format(
                _11.real, _12.real, _13.real, _11.imag, _12.imag, _13.imag))
        
        ls_out.append("*  all structure factors are in 10**-12cm")
        ls_out.append("** structure factor tensor given in crystallographic Cartesian coordinate system\n")
        return "\n".join(ls_out)

    @property
    def to_cif(self):
        ls_out = []
        if self.is_defined:
            ls_out.append("loop_")
            ls_out.append("_refln_index_h")
            ls_out.append("_refln_index_k")
            ls_out.append("_refln_index_l")
            #ls_out.append("_refln_d_spacing")
            ls_out.append("_refln_A_calc")
            ls_out.append("_refln_B_calc")
            ls_out.append("_refln_chi_11_A_calc")
            ls_out.append("_refln_chi_12_A_calc")
            ls_out.append("_refln_chi_13_A_calc")
            ls_out.append("_refln_chi_21_A_calc")
            ls_out.append("_refln_chi_22_A_calc")
            ls_out.append("_refln_chi_23_A_calc")
            ls_out.append("_refln_chi_31_A_calc")
            ls_out.append("_refln_chi_32_A_calc")
            ls_out.append("_refln_chi_33_A_calc")
            ls_out.append("_refln_chi_11_B_calc")
            ls_out.append("_refln_chi_12_B_calc")
            ls_out.append("_refln_chi_13_B_calc")
            ls_out.append("_refln_chi_21_B_calc")
            ls_out.append("_refln_chi_22_B_calc")
            ls_out.append("_refln_chi_23_B_calc")
            ls_out.append("_refln_chi_31_B_calc")
            ls_out.append("_refln_chi_32_B_calc")
            ls_out.append("_refln_chi_33_B_calc")
            for _1, _2, _3, _4, _5, _6, _7, _8, _9, _10, _11, _12, _13 in zip(self.h, self.k, self.l, self.f_nucl, 
                                                                          self.sft_11, self.sft_12, self.sft_13, 
                                                                          self.sft_21, self.sft_22, self.sft_23, 
                                                                          self.sft_31, self.sft_32, self.sft_33):
                ls_out.append("{:} {:} {:} {:.5f} {:.5f} {:.5f} {:.5f} {:.5f} {:.5f} {:.5f} {:.5f} {:.5f} {:.5f} {:.5f} {:.5f} {:.5f} {:.5f} {:.5f} {:.5f} {:.5f} {:.5f} {:.5f} {:.5f}".format(
                _1, _2, _3, _4.real, _4.imag, _5.real, _6.real, _7.real, _8.real, _9.real, _10.real, _11.real, _12.real, _13.real,
                _5.imag, _6.imag, _7.imag, _8.imag, _9.imag, _10.imag, _11.imag, _12.imag, _13.imag))
        return "\n".join(ls_out)

    def from_cif(self, string: str):
        cif_global = Global()
        flag = cif_global.take_from_string(string)
        if not flag:
            return False
        flag = False
        flag = cif_global.is_prefix("_pd_meas")
        if flag:
            cif_loop = cif_global["_pd_meas"]
            l_name = cif_loop.names
            if "_pd_meas_angle_2theta" in l_name:
                self.ttheta = [float(_1) for _1 in cif_loop["_pd_meas_angle_2theta"]]
            if "_pd_meas_intensity_up" in l_name:
                self.up = [float(_1) for _1 in cif_loop["_pd_meas_intensity_up"]]
            if "_pd_meas_intensity_up_sigma" in l_name:
                self.up_sigma = [float(_1) for _1 in cif_loop["_pd_meas_intensity_up_sigma"]]
            if "_pd_meas_intensity_down" in l_name:
                self.down = [float(_1) for _1 in cif_loop["_pd_meas_intensity_down"]]
            if "_pd_meas_intensity_down_sigma" in l_name:
                self.down_sigma = [float(_1) for _1 in cif_loop["_pd_meas_intensity_down_sigma"]]

            l_a, l_b, l_chi_11_a, l_chi_12_a, l_chi_13_a, l_chi_21_a, l_chi_22_a, l_chi_23_a = [], [], [], [], [], [], [], []
            l_chi_31_a, l_chi_32_a, l_chi_33_a = [], [], []
            l_chi_11_b, l_chi_12_b, l_chi_13_b, l_chi_21_b, l_chi_22_b, l_chi_23_b = [], [], [], [], [], []
            l_chi_31_b, l_chi_32_b, l_chi_33_b = [], [], []
            if "_refln_index_h" in l_name:
                self.h = [int(_1) for _1 in cif_loop["_refln_index_h"]]
            if "_refln_index_k" in l_name:
                self.k = [int(_1) for _1 in cif_loop["_refln_index_k"]]
            if "_refln_index_l" in l_name:
                self.l = [int(_1) for _1 in cif_loop["_refln_index_l"]]
            if "_refln_A_calc" in l_name:
                l_a = [float(_1) for _1 in cif_loop["_refln_A_calc"]]
            if "_refln_B_calc" in l_name:
                l_b = [float(_1) for _1 in cif_loop["_refln_B_calc"]]
            if "_refln_chi_11_A_calc" in l_name:
                l_chi_11_a = [float(_1) for _1 in cif_loop["_refln_chi_11_A_calc"]]
            if "_refln_chi_12_A_calc" in l_name:
                l_chi_12_a = [float(_1) for _1 in cif_loop["_refln_chi_12_A_calc"]]
            if "_refln_chi_13_A_calc" in l_name:
                l_chi_13_a = [float(_1) for _1 in cif_loop["_refln_chi_13_A_calc"]]
            if "_refln_chi_21_A_calc" in l_name:
                l_chi_21_a = [float(_1) for _1 in cif_loop["_refln_chi_21_A_calc"]]
            if "_refln_chi_22_A_calc" in l_name:
                l_chi_22_a = [float(_1) for _1 in cif_loop["_refln_chi_22_A_calc"]]
            if "_refln_chi_23_A_calc" in l_name:
                l_chi_23_a = [float(_1) for _1 in cif_loop["_refln_chi_23_A_calc"]]
            if "_refln_chi_31_A_calc" in l_name:
                l_chi_31_a = [float(_1) for _1 in cif_loop["_refln_chi_31_A_calc"]]
            if "_refln_chi_32_A_calc" in l_name:
                l_chi_32_a = [float(_1) for _1 in cif_loop["_refln_chi_32_A_calc"]]
            if "_refln_chi_33_A_calc" in l_name:
                l_chi_33_a = [float(_1) for _1 in cif_loop["_refln_chi_33_A_calc"]]
            if "_refln_chi_11_B_calc" in l_name:
                l_chi_11_b = [float(_1) for _1 in cif_loop["_refln_chi_11_B_calc"]]
            if "_refln_chi_12_B_calc" in l_name:
                l_chi_12_b = [float(_1) for _1 in cif_loop["_refln_chi_12_B_calc"]]
            if "_refln_chi_13_B_calc" in l_name:
                l_chi_13_b = [float(_1) for _1 in cif_loop["_refln_chi_13_B_calc"]]
            if "_refln_chi_21_B_calc" in l_name:
                l_chi_21_b = [float(_1) for _1 in cif_loop["_refln_chi_21_B_calc"]]
            if "_refln_chi_22_B_calc" in l_name:
                l_chi_22_b = [float(_1) for _1 in cif_loop["_refln_chi_22_B_calc"]]
            if "_refln_chi_23_B_calc" in l_name:
                l_chi_23_b = [float(_1) for _1 in cif_loop["_refln_chi_23_B_calc"]]
            if "_refln_chi_31_B_calc" in l_name:
                l_chi_31_b = [float(_1) for _1 in cif_loop["_refln_chi_31_B_calc"]]
            if "_refln_chi_32_B_calc" in l_name:
                l_chi_32_b = [float(_1) for _1 in cif_loop["_refln_chi_32_B_calc"]]
            if "_refln_chi_33_B_calc" in l_name:
                l_chi_33_b = [float(_1) for _1 in cif_loop["_refln_chi_33_B_calc"]]
            self.f_nucl = [complex(_1, _2) for _1, _2 in zip(l_a, l_b)]
            self.chi_11 = [complex(_1, _2) for _1, _2 in zip(l_chi_11_a, l_chi_11_b)]
            self.chi_12 = [complex(_1, _2) for _1, _2 in zip(l_chi_12_a, l_chi_12_b)]
            self.chi_13 = [complex(_1, _2) for _1, _2 in zip(l_chi_13_a, l_chi_13_b)]
            self.chi_21 = [complex(_1, _2) for _1, _2 in zip(l_chi_21_a, l_chi_21_b)]
            self.chi_22 = [complex(_1, _2) for _1, _2 in zip(l_chi_22_a, l_chi_22_b)]
            self.chi_23 = [complex(_1, _2) for _1, _2 in zip(l_chi_23_a, l_chi_23_b)]
            self.chi_31 = [complex(_1, _2) for _1, _2 in zip(l_chi_31_a, l_chi_31_b)]
            self.chi_32 = [complex(_1, _2) for _1, _2 in zip(l_chi_32_a, l_chi_32_b)]
            self.chi_33 = [complex(_1, _2) for _1, _2 in zip(l_chi_33_a, l_chi_33_b)]

        else:
            self.h, self.k, self.l, self.f_nucl = [], [], [], []
            self.chi_11, self.chi_12, self.chi_13 = [], [], []
            self.chi_21, self.chi_22, self.chi_23 = [], [], []
            self.chi_31, self.chi_32, self.chi_33 = [], [], []
        return True

    @property
    def is_defined(self):
        cond = all([self.h is not None, self.k is not None, self.l is not None, self.f_nucl is not None, 
                    self.sft_11 is not None, self.sft_12 is not None, self.sft_13 is not None,
                    self.sft_21 is not None, self.sft_22 is not None, self.sft_23 is not None,
                    self.sft_31 is not None, self.sft_32 is not None, self.sft_33 is not None])
        return cond

    @property
    def is_variable(self):
        return False
    
    def get_variables(self):
        return []

    def _show_message(self, s_out: str):
        print("***  Error ***")
        print(s_out)



"""
define classes to describe crystal 
"""
__author__ = 'ikibalin'
__version__ = "2019_09_02"
import os
import numpy

from pycifstar import Global
from cryspy.f_common.cl_fitable import Fitable

class Extinction(object):
    """
    Extinction
    """
    def __init__(self, radius=Fitable(0.), 
                       mosaicity=Fitable(0.), 
                       model="gauss"):
        super(Extinction, self).__init__()
        self.__refine_ls_extinction_coef_radius = None
        self.__refine_ls_extinction_coef_mosaicity = None
        self.__refine_ls_extinction_model = None
        self.model = model
        self.radius = radius
        self.mosaicity = mosaicity

    def __repr__(self):
        ls_out = []
        ls_out.append("Extinction:")
        ls_out.append("model: {:}".format(self.model))
        ls_out.append("radius: {:}".format(self.radius))
        ls_out.append("mosaicity: {:}".format(self.mosaicity))
        return "\n".join(ls_out)

    @property
    def model(self):
        """
        The model of extinction

        """
        return self.__refine_ls_extinction_model
    @model.setter
    def model(self, x):
        l_model = ["gauss", "lorentz"]
        if x is None:
            x_in = "gauss"
        else:
            x_in = str(x).strip().lower()
            if not(x_in in l_model):
                x_in = "gauss"
                self._show_message("Can not recognize the introduced model '{:}'.\nThe supported models: {}.\nThe gauss model was choosen".format(
                                    x, " ".join(l_model)))
        self.__refine_ls_extinction_model = x_in

    @property
    def radius(self):
        """
        The domain radius in angstrem
        """
        return self.__refine_ls_extinction_coef_radius
    @radius.setter
    def radius(self, x):
        if isinstance(x, Fitable):
            x_in = x
        else:
            x_in = Fitable()
            flag = x_in.take_it(x)
        self.__refine_ls_extinction_coef_radius = x_in

    @property
    def mosaicity(self):
        """
        The domains' mosaicity in angstrem
        """
        return self.__refine_ls_extinction_coef_mosaicity
    @mosaicity.setter
    def mosaicity(self, x):
        if isinstance(x, Fitable):
            x_in = x
        else:
            x_in = Fitable()
            flag = x_in.take_it(x)
        self.__refine_ls_extinction_coef_mosaicity = x_in

    def calc_extinction(self, cell, h, k, l, f_sq, wavelength, flag_derivative_f_sq=False):
        """
        f_sq in 10-12cm
        extinction for spherical model
        """
        r = 1.*self.radius
        g = 1.*self.mosaicity
        model = self.model
        kk = 1.
        vol = cell.volume
        sthovl = cell.calc_sthovl(h=h, k=k, l=l)
        stheta = sthovl * wavelength

        s2theta = 2. * stheta * (1. - stheta**2)**0.5
        c2theta = 1. - 2. * stheta**2
    
        q = (f_sq*kk/vol**2)*(wavelength**3)*1./s2theta
        delta_f_sq_q = (kk/vol**2)*(wavelength**3)*1./s2theta

        t = 1.5*r
        alpha = 1.5*r*s2theta*1./wavelength
        x = 2./3*q*alpha*t
        delta_f_sq_x = 2./3*delta_f_sq_q*alpha*t

        A = 0.20 + 0.45 * c2theta
        B = 0.22 - 0.12 * (0.5-c2theta)**2
        yp = (1.+2.*x+(A*x**2)*1./(1.+B*x))**(-0.5)
        delta_f_sq_yp = -0.5*(yp**3)*(2.*delta_f_sq_x+
                          2.*A*delta_f_sq_x*x*1./(1.+B*x)-
                          B*delta_f_sq_x*(A*x**2)*1./((1.+B*x)**2)) 
        
        
        ag = numpy.zeros(h.shape, dtype=float)
        al = numpy.zeros(h.shape, dtype=float)
        
        flag = alpha != 0.
        ag[flag] = alpha[flag]*g*(g**2+0.5*alpha[flag]**2)**(-0.5)
        al[flag] = alpha[flag]*g*1./(g+alpha[flag]*2./3.)
        
        if model == "gauss":
            xs = 2./3.*q*ag*t
            delta_f_sq_xs = 2./3.*delta_f_sq_q*ag*t
            A = 0.58 + 0.48 * c2theta + 0.24 * c2theta**2
            B = 0.02 - 0.025 * c2theta
            #print("A, B", A, B)
            ys = (1+2.12*xs+(A*xs**2)*1./(1+B*xs))**(-0.5)
            delta_f_sq_ys = -0.5*(ys**3)*(2.12*delta_f_sq_xs+2.*(A*xs*delta_f_sq_xs)*1./(1+B*xs)-B*delta_f_sq_xs*(A*xs**2)*1./((1+B*xs)**2))
        elif model == "lorentz":
            xs = 2./3.*q*al*t
            delta_f_sq_xs = 2./3.*delta_f_sq_q*al*t

            A = 0.025 + 0.285 * c2theta
            B = -0.45 * c2theta
            flag = c2theta>0
            B[flag] = 0.15 - 0.2 * (0.75-c2theta[flag])**2
            ys = (1.+2.*xs+(A*xs**2)*1./(1+B*xs))**(-0.5)
            delta_f_sq_ys = -0.5*(ys**3)*(2.*delta_f_sq_xs+2.*(A*xs*delta_f_sq_xs)*1./(1+B*xs)-B*delta_f_sq_xs*(A*xs**2)*1./((1.+B*xs)**2))
        else:
            ys = 1.
            delta_f_sq_ys = 0.
        #print("ys", ys)
        yext = yp * ys
        delta_f_sq_yext = delta_f_sq_yp * ys + yp * delta_f_sq_ys
        if flag_derivative_f_sq:
            return yext, delta_f_sq_yext 
        else:  
            return yext
            
    @property
    def is_variable(self):
        """
        without extinction
        """
        res = any([self.radius.refinement, 
                   self.mosaicity.refinement])
        return res

    def get_variables(self):
        l_variable = []
        if self.radius.refinement:
            l_variable.append(self.radius)
        if self.mosaicity.refinement:
            l_variable.append(self.mosaicity)
        return l_variable

    def _show_message(self, s_out: str):
        print("***  Error ***")
        print(s_out)

    @property
    def to_cif(self):
        ls_out = ["_refine_ls_extinction_coef_mosaicity {:}".format(self.mosaicity.print_with_sigma)]
        ls_out.append("_refine_ls_extinction_coef_radius {:}".format(self.radius.print_with_sigma))
        return "\n".join(ls_out)


    def from_cif(self, string: str):
        cif_data = Global()
        flag = cif_data.take_from_string(string)
        if not flag:
            return False
        flag = False
        if cif_data.is_value("_refine_ls_extinction_coef_mosaicity"):
            self.mosaicity = cif_data["_refine_ls_extinction_coef_mosaicity"] # CIFvalue
        if cif_data.is_value("_refine_ls_extinction_coef_radius"):
            self.radius = cif_data["_refine_ls_extinction_coef_radius"] # CIFvalue
        return True




"""
define classes to describe AtomSiteMagnetismAniso
"""
__author__ = 'ikibalin'
__version__ = "2019_08_29"
import os
import numpy


from pycifstar import Global
from cryspy.f_common.cl_fitable import Fitable

from cryspy.f_crystal.cl_atom_site_magnetism import AtomSiteMagnetism
from cryspy.f_crystal.cl_magnetism import Magnetism


class AtomSiteMagnetismAniso(object):
    """
    Data items in the ATOM_SITE_MAGNETISM_ANISO category record details about
    magnetic properties of the atoms that occupy the atom sites.
    
    Description in cif file:

    loop_  
    _atom_site_magnetism_aniso_label
    _atom_site_magnetism_aniso_chi_type
    _atom_site_magnetism_aniso_chi_11
    _atom_site_magnetism_aniso_chi_12
    _atom_site_magnetism_aniso_chi_13
    _atom_site_magnetism_aniso_chi_22
    _atom_site_magnetism_aniso_chi_23
    _atom_site_magnetism_aniso_chi_33
     Fe3A cani -3.468(74) 0.0 0.0 -3.468 0.0 -3.468
     Fe3B cani 3.041      0.0 0.0  3.041 0.0  3.041
    
    """    
    def __init__(self, label=[], chi_type=[], 
                 chi_11=[], chi_22=[], chi_33=[],
                 chi_12=[], chi_13=[], chi_23=[]):
        super(AtomSiteMagnetismAniso, self).__init__()

        self.__atom_site_magnetism_aniso_label = []
        self.__atom_site_magnetism_aniso_chi_type = []
        self.__atom_site_magnetism_aniso_chi_11 = []
        self.__atom_site_magnetism_aniso_chi_12 = []
        self.__atom_site_magnetism_aniso_chi_13 = []
        self.__atom_site_magnetism_aniso_chi_22 = []
        self.__atom_site_magnetism_aniso_chi_23 = []
        self.__atom_site_magnetism_aniso_chi_33 = []

        self.label = label
        self.chi_type = chi_type
        self.chi_11 = chi_11
        self.chi_12 = chi_12
        self.chi_13 = chi_13
        self.chi_22 = chi_22
        self.chi_23 = chi_23
        self.chi_33 = chi_33
        
    def __repr__(self):
        ls_out = ["AtomSiteMagnetismAniso:"]
        ls_out.append(" label    chi_type chi_11   chi_22   chi_33   chi_12   chi_13   chi_23   ")
        for hh_1, hh_2, hh_3, hh_4, hh_5, hh_6, hh_7, hh_8 in zip(
            self.label, self.chi_type, self.chi_11, self.chi_22, self.chi_33, self.chi_12, self.chi_13, self.chi_23):
            ls_out.append(" {:8} {:8} {:8} {:8} {:8} {:8} {:8} {:8}".format(hh_1, hh_2, 
                    hh_3.print_with_sigma, hh_4.print_with_sigma, hh_5.print_with_sigma, 
                    hh_6.print_with_sigma, hh_7.print_with_sigma, hh_8.print_with_sigma))
        return "\n".join(ls_out)

    @property
    def label(self):
        """
        The _atom_site_magnetism_aniso_label is a unique identifier for a particular site
        in the crystal. 

        Type: char
        """
        return tuple(self.__atom_site_magnetism_aniso_label)
    @label.setter
    def label(self, l_x):
        l_x_in = []
        for x in l_x:
            x_in = str(x).strip()
            l_x_in.append(x_in)
        self.__atom_site_magnetism_aniso_label = l_x_in
        len_x = len(l_x_in)

        len_1 = len(self.__atom_site_magnetism_aniso_chi_type)
        if len_1 > len_x:
            self.__atom_site_magnetism_aniso_chi_type = self.__atom_site_magnetism_aniso_chi_type[:len_x]
        elif len_1 < len_x:
            l_fitable = ["Fe3+" for hh in range(len_x-len_1)]
            self.__atom_site_magnetism_aniso_chi_type.extend(l_fitable)

        len_1 = len(self.__atom_site_magnetism_aniso_chi_11)
        if len_1 > len_x:
            self.__atom_site_magnetism_aniso_chi_11 = self.__atom_site_magnetism_aniso_chi_11[:len_x]
        elif len_1 < len_x:
            l_fitable = [Fitable(value=0., name="_atom_site_magnetism_aniso_chi_11") for hh in range(len_x-len_1)]
            self.__atom_site_magnetism_aniso_chi_11.extend(l_fitable)

        len_1 = len(self.__atom_site_magnetism_aniso_chi_22)
        if len_1 > len_x:
            self.__atom_site_magnetism_aniso_chi_22 = self.__atom_site_magnetism_aniso_chi_22[:len_x]
        elif len_1 < len_x:
            l_fitable = [Fitable(value=0., name="_atom_site_magnetism_aniso_chi_22") for hh in range(len_x-len_1)]
            self.__atom_site_magnetism_aniso_chi_22.extend(l_fitable)

        len_1 = len(self.__atom_site_magnetism_aniso_chi_33)
        if len_1 > len_x:
            self.__atom_site_magnetism_aniso_chi_33 = self.__atom_site_magnetism_aniso_chi_33[:len_x]
        elif len_1 < len_x:
            l_fitable = [Fitable(value=0., name="_atom_site_magnetism_aniso_chi_33") for hh in range(len_x-len_1)]
            self.__atom_site_magnetism_aniso_chi_33.extend(l_fitable)

        len_1 = len(self.__atom_site_magnetism_aniso_chi_11)
        if len_1 > len_x:
            self.__atom_site_magnetism_aniso_chi_12 = self.__atom_site_magnetism_aniso_chi_12[:len_x]
        elif len_1 < len_x:
            l_fitable = [Fitable(value=0., name="_atom_site_magnetism_aniso_chi_12") for hh in range(len_x-len_1)]
            self.__atom_site_magnetism_aniso_chi_12.extend(l_fitable)

        len_1 = len(self.__atom_site_magnetism_aniso_chi_13)
        if len_1 > len_x:
            self.__atom_site_magnetism_aniso_chi_13 = self.__atom_site_magnetism_aniso_chi_13[:len_x]
        elif len_1 < len_x:
            l_fitable = [Fitable(value=0., name="_atom_site_magnetism_aniso_chi_13") for hh in range(len_x-len_1)]
            self.__atom_site_magnetism_aniso_chi_13.extend(l_fitable)

        len_1 = len(self.__atom_site_magnetism_aniso_chi_23)
        if len_1 > len_x:
            self.__atom_site_magnetism_aniso_chi_23 = self.__atom_site_magnetism_aniso_chi_23[:len_x]
        elif len_1 < len_x:
            l_fitable = [Fitable(value=0., name="_atom_site_magnetism_aniso_chi_23") for hh in range(len_x-len_1)]
            self.__atom_site_magnetism_aniso_chi_23.extend(l_fitable)

    @property
    def chi_type(self):
        """
        Chi type. 

        Default: Fe3+

        Type: char
        """
        return tuple(self.__atom_site_magnetism_aniso_chi_type)
    @chi_type.setter
    def chi_type(self, l_x):
        l_fitable = []
        for x in l_x:
            x_in = str(x).strip()
            l_fitable.append(x_in)
        len_x = len(l_fitable)
        len_1 = len(self.__atom_site_magnetism_aniso_label)
        if len_1 < len_x:
            l_fitable = l_fitable[:len_1]
        elif len_1 > len_x:
            l_fitable.extend(["Fe3+" for hh in range(len_1-len_x)])
        self.__atom_site_magnetism_aniso_chi_type = l_fitable

    @property
    def chi_11(self):
        """
        chi_11

        Default: 0.

        Type: float
        """
        return tuple(self.__atom_site_magnetism_aniso_chi_11)
    @chi_11.setter
    def chi_11(self, l_x):
        l_fitable = []
        for x in l_x:
            if isinstance(x, Fitable):
                x_in = x
            else:
                x_in = Fitable()
                flag = x_in.take_it(x)
            l_fitable.append(x_in)
        len_x = len(l_fitable)
        len_1 = len(self.__atom_site_magnetism_aniso_label)
        if len_1 < len_x:
            l_fitable = l_fitable[:len_1]
        elif len_1 > len_x:
            l_fitable.extend([Fitable(value=0., name= "_atom_site_magnetism_aniso_chi_11") for hh in range(len_1-len_x)])
        self.__atom_site_magnetism_aniso_chi_11 = l_fitable


    @property
    def chi_22(self):
        """
        chi_22

        Default: 0.

        Type: float
        """
        return tuple(self.__atom_site_magnetism_aniso_chi_22)
    @chi_22.setter
    def chi_22(self, l_x):
        l_fitable = []
        for x in l_x:
            if isinstance(x, Fitable):
                x_in = x
            else:
                x_in = Fitable()
                flag = x_in.take_it(x)
            l_fitable.append(x_in)
        len_x = len(l_fitable)
        len_1 = len(self.__atom_site_magnetism_aniso_label)
        if len_1 < len_x:
            l_fitable = l_fitable[:len_1]
        elif len_1 > len_x:
            l_fitable.extend([Fitable(value=0., name= "_atom_site_magnetism_aniso_chi_22") for hh in range(len_1-len_x)])
        self.__atom_site_magnetism_aniso_chi_22 = l_fitable


    @property
    def chi_33(self):
        """
        chi_33

        Default: 0.

        Type: float
        """
        return tuple(self.__atom_site_magnetism_aniso_chi_33)
    @chi_33.setter
    def chi_33(self, l_x):
        l_fitable = []
        for x in l_x:
            if isinstance(x, Fitable):
                x_in = x
            else:
                x_in = Fitable()
                flag = x_in.take_it(x)
            l_fitable.append(x_in)
        len_x = len(l_fitable)
        len_1 = len(self.__atom_site_magnetism_aniso_label)
        if len_1 < len_x:
            l_fitable = l_fitable[:len_1]
        elif len_1 > len_x:
            l_fitable.extend([Fitable(value=0., name= "_atom_site_magnetism_aniso_chi_33") for hh in range(len_1-len_x)])
        self.__atom_site_magnetism_aniso_chi_33 = l_fitable


    @property
    def chi_12(self):
        """
        chi_12

        Default: 0.

        Type: float
        """
        return tuple(self.__atom_site_magnetism_aniso_chi_12)
    @chi_12.setter
    def chi_12(self, l_x):
        l_fitable = []
        for x in l_x:
            if isinstance(x, Fitable):
                x_in = x
            else:
                x_in = Fitable()
                flag = x_in.take_it(x)
            l_fitable.append(x_in)
        len_x = len(l_fitable)
        len_1 = len(self.__atom_site_magnetism_aniso_label)
        if len_1 < len_x:
            l_fitable = l_fitable[:len_1]
        elif len_1 > len_x:
            l_fitable.extend([Fitable(value=0., name= "_atom_site_magnetism_aniso_chi_12") for hh in range(len_1-len_x)])
        self.__atom_site_magnetism_aniso_chi_12 = l_fitable


    @property
    def chi_13(self):
        """
        chi_13

        Default: 0.

        Type: float
        """
        return tuple(self.__atom_site_magnetism_aniso_chi_13)
    @chi_13.setter
    def chi_13(self, l_x):
        l_fitable = []
        for x in l_x:
            if isinstance(x, Fitable):
                x_in = x
            else:
                x_in = Fitable()
                flag = x_in.take_it(x)
            l_fitable.append(x_in)
        len_x = len(l_fitable)
        len_1 = len(self.__atom_site_magnetism_aniso_label)
        if len_1 < len_x:
            l_fitable = l_fitable[:len_1]
        elif len_1 > len_x:
            l_fitable.extend([Fitable(value=0., name= "_atom_site_magnetism_aniso_chi_13") for hh in range(len_1-len_x)])
        self.__atom_site_magnetism_aniso_chi_13 = l_fitable


    @property
    def chi_23(self):
        """
        chi_23

        Default: 0.

        Type: float
        """
        return tuple(self.__atom_site_magnetism_aniso_chi_23)
    @chi_23.setter
    def chi_23(self, l_x):
        l_fitable = []
        for x in l_x:
            if isinstance(x, Fitable):
                x_in = x
            else:
                x_in = Fitable()
                flag = x_in.take_it(x)
            l_fitable.append(x_in)
        len_x = len(l_fitable)
        len_1 = len(self.__atom_site_magnetism_aniso_label)
        if len_1 < len_x:
            l_fitable = l_fitable[:len_1]
        elif len_1 > len_x:
            l_fitable.extend([Fitable(value=0., name= "_atom_site_magnetism_aniso_chi_23") for hh in range(len_1-len_x)])
        self.__atom_site_magnetism_aniso_chi_23 = l_fitable

    def _show_message(self, s_out: str):
        print("***  Error ***")
        print(s_out)

    @property
    def is_defined(self):
        """
        Output: True if all started parameters are given
        """
        cond = (self.label != [])
        return cond
    @property
    def is_variable(self):
        res = (any([hh.refinement for hh in self.chi_11]) | 
               any([hh.refinement for hh in self.chi_22]) |
               any([hh.refinement for hh in self.chi_33]) |
               any([hh.refinement for hh in self.chi_12]) |
               any([hh.refinement for hh in self.chi_13]) |
               any([hh.refinement for hh in self.chi_23]))
        return res


    def get_variables(self):
        l_variable = [hh for hh in self.chi_11 if hh.refinement]
        l_variable.extend([hh for hh in self.chi_22 if hh.refinement])
        l_variable.extend([hh for hh in self.chi_33 if hh.refinement])
        l_variable.extend([hh for hh in self.chi_12 if hh.refinement])
        l_variable.extend([hh for hh in self.chi_13 if hh.refinement])
        l_variable.extend([hh for hh in self.chi_23 if hh.refinement])

        return l_variable

    @property
    def to_cif(self):
        ls_out = []
        if self.is_defined:
            ls_out.append("loop_")
            ls_out.append("_atom_site_magnetism_aniso_label")
            ls_out.append("_atom_site_magnetism_aniso_chi_type")
            ls_out.append("_atom_site_magnetism_aniso_chi_11")
            ls_out.append("_atom_site_magnetism_aniso_chi_22")
            ls_out.append("_atom_site_magnetism_aniso_chi_33")
            ls_out.append("_atom_site_magnetism_aniso_chi_12")
            ls_out.append("_atom_site_magnetism_aniso_chi_13")
            ls_out.append("_atom_site_magnetism_aniso_chi_23")
            for hh_1, hh_2, hh_3, hh_4, hh_5, hh_6, hh_7, hh_8 in zip(self.label, self.chi_type, 
                self.chi_11, self.chi_22, self.chi_33, self.chi_12, self.chi_13, self.chi_23):
                ls_out.append("{:} {:} {:} {:} {:} {:} {:} {:}".format(hh_1, hh_2, 
                        hh_3.print_with_sigma, hh_4.print_with_sigma, hh_5.print_with_sigma,
                        hh_6.print_with_sigma, hh_7.print_with_sigma, hh_8.print_with_sigma))
        return "\n".join(ls_out)

    def from_cif(self, string: str):
        cif_global = Global()
        flag = cif_global.take_from_string(string)
        if not flag:
            return False
        flag = False
        flag = cif_global.is_prefix("_atom_site_magnetism_aniso")
        if flag:
            cif_loop = cif_global["_atom_site_magnetism_aniso"]
            l_name = cif_loop.names
            if "_atom_site_magnetism_aniso_label" in l_name:
                self.label = cif_loop["_atom_site_magnetism_aniso_label"]

            if "_atom_site_magnetism_aniso_chi_type" in l_name:
                self.chi_type = cif_loop["_atom_site_magnetism_aniso_chi_type"]
            if "_atom_site_magnetism_aniso_chi_11" in l_name:
                l_val = cif_loop["_atom_site_magnetism_aniso_chi_11"]
                l_fitable = []
                for val in l_val:
                    fitable = Fitable(name="_atom_site_magnetism_aniso_chi_11")
                    fitable.take_it(val)
                    l_fitable.append(fitable)
                self.chi_11 = l_fitable
            if "_atom_site_magnetism_aniso_chi_22" in l_name:
                l_val = cif_loop["_atom_site_magnetism_aniso_chi_22"]
                l_fitable = []
                for val in l_val:
                    fitable = Fitable(name="_atom_site_magnetism_aniso_chi_22")
                    fitable.take_it(val)
                    l_fitable.append(fitable)
                self.chi_22 = l_fitable
            if "_atom_site_magnetism_aniso_chi_33" in l_name:
                l_val = cif_loop["_atom_site_magnetism_aniso_chi_33"]
                l_fitable = []
                for val in l_val:
                    fitable = Fitable(name="_atom_site_magnetism_aniso_chi_33")
                    fitable.take_it(val)
                    l_fitable.append(fitable)
                self.chi_33 = l_fitable
            if "_atom_site_magnetism_aniso_chi_12" in l_name:
                l_val = cif_loop["_atom_site_magnetism_aniso_chi_12"]
                l_fitable = []
                for val in l_val:
                    fitable = Fitable(name="_atom_site_magnetism_aniso_chi_12")
                    fitable.take_it(val)
                    l_fitable.append(fitable)
                self.chi_12 = l_fitable
            if "_atom_site_magnetism_aniso_chi_13" in l_name:
                l_val = cif_loop["_atom_site_magnetism_aniso_chi_13"]
                l_fitable = []
                for val in l_val:
                    fitable = Fitable(name="_atom_site_magnetism_aniso_chi_13")
                    fitable.take_it(val)
                    l_fitable.append(fitable)
                self.chi_13 = l_fitable
            if "_atom_site_magnetism_aniso_chi_23" in l_name:
                l_val = cif_loop["_atom_site_magnetism_aniso_chi_23"]
                l_fitable = []
                for val in l_val:
                    fitable = Fitable(name="_atom_site_magnetism_aniso_chi_23")
                    fitable.take_it(val)
                    l_fitable.append(fitable)
                self.chi_23 = l_fitable
        else:
            self.label = []

        return True

    def _form_magnetism(self, atom_site, atom_magnetism):
        label = numpy.array(atom_site.label, dtype=str)
        label_aniso = numpy.array(self.label, dtype=str)
        if not(set(label_aniso).issubset(set(label))):
            self._show_message("Unknown 'aniso_label'")
            return False
        
        j0_A_in, j2_A_in = atom_site.j0_A, atom_site.j2_A
        j0_a_in, j2_a_in = atom_site.j0_a, atom_site.j2_a
        j0_B_in, j2_B_in = atom_site.j0_B, atom_site.j2_B
        j0_b_in, j2_b_in = atom_site.j0_b, atom_site.j2_b
        j0_C_in, j2_C_in = atom_site.j0_C, atom_site.j2_C
        j0_c_in, j2_c_in = atom_site.j0_c, atom_site.j2_c
        j0_D_in, j2_D_in = atom_site.j0_D, atom_site.j2_D

        lande_in, kappa_in = atom_magnetism._form_lande_kappa(atom_site)
        
        #chi_type = numpy.array(self.chi_type, dtype=str)
        chi_11 = numpy.array(self.chi_11, dtype=float)
        chi_22 = numpy.array(self.chi_22, dtype=float)
        chi_33 = numpy.array(self.chi_33, dtype=float)
        chi_12 = numpy.array(self.chi_12, dtype=float)
        chi_13 = numpy.array(self.chi_13, dtype=float)
        chi_23 = numpy.array(self.chi_23, dtype=float)


        np_index = numpy.array([int(numpy.argwhere(label==hh)[0]) for hh in label_aniso], dtype=int)

        chi_11_in = numpy.zeros(label.shape, dtype=float)
        chi_22_in = numpy.zeros(label.shape, dtype=float)
        chi_33_in = numpy.zeros(label.shape, dtype=float)
        chi_12_in = numpy.zeros(label.shape, dtype=float)
        chi_13_in = numpy.zeros(label.shape, dtype=float)
        chi_23_in = numpy.zeros(label.shape, dtype=float)

        chi_11_in[np_index], chi_22_in[np_index], chi_33_in[np_index] = chi_11, chi_22, chi_33
        chi_12_in[np_index], chi_13_in[np_index], chi_23_in[np_index] = chi_12, chi_13, chi_23



        magnetism = Magnetism(factor_lande=lande_in, kappa=kappa_in, 
                              chi_11=chi_11_in, chi_22=chi_22_in, chi_33=chi_33_in,
                              chi_12=chi_12_in, chi_13=chi_13_in, chi_23=chi_23_in,
                              j0_A=j0_A_in, j0_a=j0_a_in, j0_B=j0_B_in, j0_b=j0_b_in,
                              j0_C=j0_C_in, j0_c=j0_c_in, j0_D=j0_D_in,
                              j2_A=j2_A_in, j2_a=j2_a_in, j2_B=j2_B_in, j2_b=j2_b_in,
                              j2_C=j2_C_in, j2_c=j2_c_in, j2_D=j2_D_in)

        return magnetism

    def apply_space_group_constraint(self, atom_site, space_group):
        """
        according to table 1 in Peterse, Palm, Acta Cryst.(1966), 20, 147
        """
        l_numb = atom_site.calc_constr_number(space_group)
        label_aniso = self.label
        label = atom_site.label
        l_ind = [label.index(_1) for _1 in label_aniso]
        for index, chi_11, chi_22, chi_33, chi_12, chi_13, chi_23 in zip(l_ind, self.chi_11, self.chi_22, self.chi_33, 
                                                                                self.chi_12, self.chi_13, self.chi_23):
            chi_11.constraint_flag, chi_22.constraint_flag, chi_33.constraint_flag = False, False, False
            chi_12.constraint_flag, chi_13.constraint_flag, chi_23.constraint_flag = False, False, False
            numb = l_numb[index]
            if numb == 1:
                chi_12.value = 0.
                chi_12.refinement = False
                chi_12.constraint_flag = True
                chi_23.value = 0.
                chi_23.refinement = False
                chi_23.constraint_flag = True
            elif numb == 2:
                chi_23.value = 0.
                chi_23.refinement = False
                chi_23.constraint_flag = True
                chi_13.value = 0.
                chi_13.refinement = False
                chi_13.constraint_flag = True
            elif numb == 3:
                chi_12.value = 0.
                chi_12.refinement = False
                chi_12.constraint_flag = True
                chi_13.value = 0.
                chi_13.refinement = False
                chi_13.constraint_flag = True
            elif numb == 4:
                chi_12.value = 0.
                chi_12.refinement = False
                chi_12.constraint_flag = True
                chi_13.value = 0.
                chi_13.refinement = False
                chi_13.constraint_flag = True
                chi_23.value = 0.
                chi_23.refinement = False
                chi_23.constraint_flag = True
            elif numb == 5:
                chi_22.value = chi_11.value
                chi_22.refinement = False
                chi_22.constraint_flag = True
                chi_13.value = 0.
                chi_13.refinement = False
                chi_13.constraint_flag = True
                chi_23.value = 0.
                chi_23.refinement = False
                chi_23.constraint_flag = True
            elif numb == 6:
                chi_22.value = chi_11.value
                chi_22.refinement = False
                chi_22.constraint_flag = True
                chi_23.value = chi_13.value 
                chi_23.refinement = False
                chi_23.constraint_flag = True
            elif numb == 7:
                chi_22.value = chi_11.value
                chi_22.refinement = False
                chi_22.constraint_flag = True
                chi_23.value = -1.*chi_13.value 
                chi_23.refinement = False
                chi_23.constraint_flag = True
            elif numb == 8:
                chi_22.value = chi_11.value
                chi_22.refinement = False
                chi_22.constraint_flag = True
                chi_12.value = 0.
                chi_12.refinement = False
                chi_12.constraint_flag = True
                chi_13.value = 0.
                chi_13.refinement = False
                chi_13.constraint_flag = True
                chi_23.value = 0.
                chi_23.refinement = False
                chi_23.constraint_flag = True
            elif numb == 9:
                chi_33.value = chi_22.value
                chi_33.refinement = False
                chi_33.constraint_flag = True
                chi_12.value = 0.
                chi_12.refinement = False
                chi_12.constraint_flag = True
                chi_13.value = 0.
                chi_13.refinement = False
                chi_13.constraint_flag = True
            elif numb == 10:
                chi_33.value = chi_22.value
                chi_33.refinement = False
                chi_33.constraint_flag = True
                chi_13.value = 1.*chi_12.value 
                chi_13.refinement = False
                chi_13.constraint_flag = True
            elif numb == 11:
                chi_33.value = chi_22.value
                chi_33.refinement = False
                chi_33.constraint_flag = True
                chi_13.value = -1.*chi_12.value 
                chi_13.refinement = False
                chi_13.constraint_flag = True
            elif numb == 12:
                chi_33.value = chi_22.value
                chi_33.refinement = False
                chi_33.constraint_flag = True
                chi_12.value = 0.
                chi_12.refinement = False
                chi_12.constraint_flag = True
                chi_13.value = 0.
                chi_13.refinement = False
                chi_13.constraint_flag = True
                chi_23.value = 0.
                chi_23.refinement = False
                chi_23.constraint_flag = True
            elif numb == 13:
                chi_12.value = 0.5*chi_22.value
                chi_12.refinement = False
                chi_12.constraint_flag = True
                chi_23.value = 0.
                chi_23.refinement = False
                chi_23.constraint_flag = True
            elif numb == 14:
                chi_12.value = 0.5*chi_22.value
                chi_12.refinement = False
                chi_12.constraint_flag = True
                chi_13.value = 0.
                chi_13.refinement = False
                chi_13.constraint_flag = True
                chi_23.value = 0.
                chi_23.refinement = False
                chi_23.constraint_flag = True
            elif numb == 15:
                chi_12.value = 0.5*chi_22.value
                chi_12.refinement = False
                chi_12.constraint_flag = True
                chi_23.value = 2.*chi_13.value
                chi_23.refinement = False
                chi_23.constraint_flag = True
            elif numb == 16:
                chi_22.value = 1.0*chi_11.value
                chi_22.refinement = False
                chi_22.constraint_flag = True
                chi_12.value = 0.5*chi_11.value
                chi_12.refinement = False
                chi_12.constraint_flag = True
                chi_13.value = 0.
                chi_13.refinement = False
                chi_13.constraint_flag = True
                chi_23.value = 0.
                chi_23.refinement = False
                chi_23.constraint_flag = True
            elif numb == 17:
                chi_22.value = 1.0*chi_11.value
                chi_22.refinement = False
                chi_22.constraint_flag = True
                chi_33.value = 1.0*chi_11.value
                chi_33.refinement = False
                chi_33.constraint_flag = True
                chi_12.value = 0.
                chi_12.refinement = False
                chi_12.constraint_flag = True
                chi_13.value = 0.
                chi_13.refinement = False
                chi_13.constraint_flag = True
                chi_23.value = 0.
                chi_23.refinement = False
                chi_23.constraint_flag = True
            elif numb == 18:
                chi_22.value = 1.0*chi_11.value
                chi_22.refinement = False
                chi_22.constraint_flag = True
                chi_33.value = 1.0*chi_11.value
                chi_33.refinement = False
                chi_33.constraint_flag = True
                chi_13.value = 1.0*chi_12.value
                chi_13.refinement = False
                chi_13.constraint_flag = True
                chi_23.value = 1.0*chi_12.value
                chi_23.refinement = False
                chi_23.constraint_flag = True


    def apply_chi_iso_constraint(self, cell):
        c_a = cell.cos_a
        s_ib = cell.sin_ib
        s_ig = cell.sin_ig
        c_ib = cell.cos_ib
        c_ig = cell.cos_ig
        #not sure, it is better to check
        for chi_type, chi_11, chi_22, chi_33, chi_12, chi_13, chi_23 in zip(self.chi_type, self.chi_11, self.chi_22, self.chi_33, 
                                                                                           self.chi_12, self.chi_13, self.chi_23):
            if chi_type.lower().startswith("ciso"):
                chi_22.value = chi_11.value
                chi_33.value = chi_11.value
                chi_12.value = chi_11.value*c_ig
                chi_13.value = chi_11.value*c_ib
                chi_23.value = chi_11.value*(c_ib*c_ig-s_ib*s_ig*c_a)
                chi_22.refinement, chi_33.refinement, chi_12.refinement, chi_13.refinement, chi_23.refinement =False, False, False, False, False
        return chi_11, chi_22, chi_33, chi_12, chi_13, chi_23


"""
internal class to calculate Debye-Waller factor
"""
__author__ = 'ikibalin'
__version__ = "2019_08_29"
import os
import numpy


class ADP(object):
    """
    ADP
    """
    def __init__(self, u_11=numpy.array([0], dtype=float), 
                       u_22=numpy.array([0], dtype=float), 
                       u_33=numpy.array([0], dtype=float), 
                       u_12=numpy.array([0], dtype=float), 
                       u_13=numpy.array([0], dtype=float), 
                       u_23=numpy.array([0], dtype=float), 
                       b_iso=numpy.array([0], dtype=float)):
        super(ADP, self).__init__()
        self.__atom_site_aniso_u_11 = None
        self.__atom_site_aniso_u_22 = None
        self.__atom_site_aniso_u_33 = None
        self.__atom_site_aniso_u_12 = None
        self.__atom_site_aniso_u_13 = None
        self.__atom_site_aniso_u_23 = None
        self.__atom_site_b_iso_or_equiv = None

        self.u_11 = u_11
        self.u_22 = u_22
        self.u_33 = u_33
        self.u_12 = u_12
        self.u_13 = u_13
        self.u_23 = u_23
        self.b_iso = b_iso

    def _trans_to_float_array(self, x):
        if isinstance(x, numpy.ndarray):
            x_out = x.astype(float)
        else:
            x_out = numpy.array([x], dtype=float)
        return x_out
        
    @property
    def u_11(self):
        """

        reference:
        """
        return self.__atom_site_aniso_u_11
    @u_11.setter
    def u_11(self, x):
        self.__atom_site_aniso_u_11 = self._trans_to_float_array(x)

    @property
    def u_22(self):
        """

        reference:
        """
        return self.__atom_site_aniso_u_22
    @u_22.setter
    def u_22(self, x):
        self.__atom_site_aniso_u_22 = self._trans_to_float_array(x)

    @property
    def u_33(self):
        """

        reference:
        """
        return self.__atom_site_aniso_u_33
    @u_33.setter
    def u_33(self, x):
        self.__atom_site_aniso_u_33 = self._trans_to_float_array(x)

    @property
    def u_12(self):
        """

        reference:
        """
        return self.__atom_site_aniso_u_12
    @u_12.setter
    def u_12(self, x):
        self.__atom_site_aniso_u_12 = self._trans_to_float_array(x)

    @property
    def u_13(self):
        """

        reference:
        """
        return self.__atom_site_aniso_u_13
    @u_13.setter
    def u_13(self, x):
        self.__atom_site_aniso_u_13 = self._trans_to_float_array(x)

    @property
    def u_23(self):
        """

        reference:
        """
        return self.__atom_site_aniso_u_23
    @u_23.setter
    def u_23(self, x):
        self.__atom_site_aniso_u_23 = self._trans_to_float_array(x)


    @property
    def b_iso(self):
        """

        reference:
        """
        return self.__atom_site_b_iso_or_equiv
    @b_iso.setter
    def b_iso(self, x):
        self.__atom_site_b_iso_or_equiv = self._trans_to_float_array(x)



    def __repr__(self):
        ls_out = ["Debye Waller:\n   U_iso    U_11     U_22     U_33    U_12     U_13     U_23"]
        ls_out.extend(["{:8.5f} {:8.5f} {:8.5f}{:8.5f} {:8.5f} {:8.5f} {:8.5f}".format(
            hh_1, hh_2, hh_3, hh_4, hh_5, hh_6, hh_7) for hh_1, hh_2, hh_3, hh_4, hh_5, hh_6, hh_7 
            in zip(self.b_iso, self.u_11, self.u_22, self.u_33, self.u_12, self.u_13, self.u_23)])
        return "\n".join(ls_out)

    def _calc_power_dwf_iso(self, sthovl):
        """
        isotropic harmonic Debye-Waller factor
        """
        b_iso = 1.*self.b_iso
        sthovl_sq = sthovl**2
        b_iso_2d, sthovl_sq_2d = numpy.meshgrid(sthovl_sq, b_iso, indexing="ij")
        
        power_dwf_iso_2d = b_iso_2d*sthovl_sq_2d
        return power_dwf_iso_2d

    def calc_power_dwf_aniso(self, space_group, cell, h, k, l):
        """
        anisotropic harmonic Debye-Waller factor
        
        h,k,l is 1D (temporary solution)
        """
        r_11, r_12 = space_group.r_11, space_group.r_12
        r_13, r_21 = space_group.r_13, space_group.r_21
        r_22, r_23 = space_group.r_22, space_group.r_23
        r_31, r_32 = space_group.r_31, space_group.r_32
        r_33 = space_group.r_33
  
        b_11, b_22, b_33, b_12, b_13, b_23 = self.calc_beta(cell)

        np_h, np_b_11, np_r_11 = numpy.meshgrid(h, b_11, r_11, indexing="ij")
        np_k, np_b_22, np_r_22 = numpy.meshgrid(k, b_22, r_22, indexing="ij")
        np_l, np_b_33, np_r_33 = numpy.meshgrid(l, b_33, r_33, indexing="ij")
        np_h, np_b_12, np_r_12 = numpy.meshgrid(h, b_12, r_12, indexing="ij")
        np_h, np_b_13, np_r_13 = numpy.meshgrid(h, b_13, r_13, indexing="ij")
        np_h, np_b_23, np_r_23 = numpy.meshgrid(h, b_23, r_23, indexing="ij")
        np_r_21 = numpy.meshgrid(h, b_23, r_21, indexing="ij")[2]
        np_r_31 = numpy.meshgrid(h, b_23, r_31, indexing="ij")[2]
        np_r_32 = numpy.meshgrid(h, b_23, r_32, indexing="ij")[2]
        
        np_h_s = np_h*np_r_11 + np_k*np_r_21 + np_l*np_r_31
        np_k_s = np_h*np_r_12 + np_k*np_r_22 + np_l*np_r_32
        np_l_s = np_h*np_r_13 + np_k*np_r_23 + np_l*np_r_33
        
        power_dwf_aniso = (np_b_11*np_h_s**2 + np_b_22*np_k_s**2 + 
                       np_b_33*np_l_s**2 + 2.*np_b_12*np_h_s*np_k_s + 
                       2.*np_b_13*np_h_s*np_l_s + 2.*np_b_23*np_k_s*np_l_s)
        
        return power_dwf_aniso 
        
    def calc_dwf(self, space_group, cell, h, k, l):
        """
        calculate Debye-Waller factor
        """
        sthovl = cell.calc_sthovl(h, k, l)
        #dimensions (hkl, atoms in assymmetric unit cell)
        power_iso_2d = self._calc_power_dwf_iso(sthovl)
        #dimensions (hkl, atoms in assymmetric unit cell, el.symmetry)
        power_aniso_3d = self.calc_power_dwf_aniso(space_group, cell, h, k, l)
        power_3d = power_iso_2d[:, :, numpy.newaxis] + power_aniso_3d
        dwf_3d = numpy.exp(-power_3d)
        return dwf_3d
        
    def calc_beta(self, cell):
        """
        calculate beta_ij from U_ij
        """
        ia, ib, ic = 1.*cell.ia, 1.*cell.ib, 1.*cell.ic
        u_11, u_22, u_33 = 1.*self.u_11, 1.*self.u_22, 1.*self.u_33
        u_12, u_13, u_23 = 1.*self.u_12, 1.*self.u_13, 1.*self.u_23
        
        beta_11 = 2.*numpy.pi**2 * u_11 *ia**2
        beta_22 = 2.*numpy.pi**2 * u_22 *ib**2
        beta_33 = 2.*numpy.pi**2 * u_33 *ic**2
        beta_12 = 2.*numpy.pi**2 * u_12 *ia*ib
        beta_13 = 2.*numpy.pi**2 * u_13 *ia*ic
        beta_23 = 2.*numpy.pi**2 * u_23 *ib*ic
        return beta_11, beta_22, beta_33, beta_12, beta_13, beta_23

